import argparse
import os
import socket
import time
from typing import Optional

import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import TimeoutLock

from socs.agents.scpi_psu.drivers import PsuInterface, ScpiPsuInterface

# For logging
txaio.use_twisted()


class ScpiPsuAgent:
    def __init__(self, agent, ip_address, gpib_slot=None, port=None):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.job = None
        self.ip_address = ip_address
        self.gpib_slot = None
        self.port = None
        self.monitor = False

        self.psu: Optional[ScpiPsuAgent] = None

        if gpib_slot is not None:
            self.gpib_slot = int(gpib_slot)
        if port is not None:
            self.port = int(port)

        # Registers Temperature and Voltage feeds
        agg_params = {
            'frame_length': 10 * 60,
        }
        self.agent.register_feed('psu_output',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    def init(self, session, params=None):
        """init()

        **Task** - Initialize connection to the power supply.

        """
        with self.lock.acquire_timeout(0) as acquired:
            if not acquired:
                return False, "Could not acquire lock"

            if self.gpib_slot is None and self.port is None:
                self.log.error('Either --gpib-slot or --port must be specified')
                return False, "Parameter not set"
            elif self.port is None:  # Use the old Prologix-based GPIB code
                try:
                    self.psu = PsuInterface(self.ip_address, self.gpib_slot)
                    self.idn = self.psu.identify()
                except socket.timeout as e:
                    self.log.error(f"PSU timed out during connect: {e}")
                    return False, "Timeout"
            else:  # Use the new direct ethernet connection code
                try:
                    self.psu = ScpiPsuInterface(self.ip_address, port=self.port)
                    self.idn = self.psu.identify()
                except socket.timeout as e:
                    self.log.error(f"PSU timed out during connect: {e}")
                    return False, "Timeout"
                except ValueError as e:
                    if (e.args[0].startswith('Model number')):
                        self.log.warn(f"PSU initialization: {e}. \
                                Number of channels defaults to 3. \
                                Suggest appending {e.args[-1]} to the list \
                                of known model numbers in scpi_psu/drivers.py")
                    else:
                        self.log.error(f"PSU initialization resulted in unknown ValueError: {e}")
                        return False, "ValueError"

            self.log.info("Connected to psu: {}".format(self.idn))

        auto_acquire = params.get('auto_acquire', False)

        if auto_acquire:
            acq_params = None
            if self.psu.num_channels != 0:
                acq_params = {'channels': [i + 1 for i in range(self.psu.num_channels)]}
            self.agent.start('monitor_output', acq_params)
        return True, 'Initialized PSU.'

    def _initialize_module(self):
        """Initialize the ScpiPsu module."""
        try:
            if self.port is None:
                self.psu = PsuInterface(self.ip_address, self.gpib_slot)
            else:
                self.psu = ScpiPsuInterface(self.ip_address, port=self.port)
        except (socket.timeout, OSError) as e:
            self.log.warn(f"Error establishing connection: {e}")
            self.psu = None
            return False

        self.idn = self.psu.identify()
        self.log.info("Connected to psu: {}".format(self.idn))
        self.log.info("Clearing event registers and error queue")
        self.psu.clear()
        return True

    @ocs_agent.param('wait', type=float, default=1)
    @ocs_agent.param('channels', type=list, default=[1, 2, 3])
    @ocs_agent.param('test_mode', type=bool, default=False)
    def monitor_output(self, session, params=None):
        """monitor_output(wait=1, channels=[1, 2, 3], test_mode=False)

        **Process** - Continuously monitor PSU output current and voltage.

        Parameters:
            wait (float, optional): Time to wait between measurements
                [seconds].
            channels (list[int], optional): Channels to monitor. [1, 2, 3] by
                default.
            test_mode (bool, optional): Exit process after single loop if True.
                Defaults to False.

        """
        self.monitor = True

        while self.monitor:
            time.sleep(params['wait'])
            with self.lock.acquire_timeout(1) as acquired:
                if not acquired:
                    self.log.warn("Could not acquire in monitor_current")
                    continue

                if not self.psu:
                    self._initialize_module()
                    continue

                data = {
                    'timestamp': time.time(),
                    'block_name': 'output',
                    'data': {}
                }

                try:
                    for chan in params['channels']:
                        if self.psu.get_output(chan):
                            data['data']["Voltage_{}".format(chan)] = self.psu.get_volt(chan)
                            data['data']["Current_{}".format(chan)] = self.psu.get_curr(chan)
                        else:
                            self.log.warn("Cannot measure output when output is disabled")
                            self.monitor = False
                            return False, "Cannot measure output when output is disabled"
                except socket.timeout as e:
                    self.log.warn(f"TimeoutError: {e}")
                    self.log.info("Attempting to reconnect")
                    self.psu = None
                    continue

                self.agent.publish_to_feed('psu_output', data)

                # Allow this process to be queried to return current data
                session.data = data

            if params['test_mode']:
                break

        return True, "Finished monitoring current"

    def stop_monitoring(self, session, params=None):
        self.monitor = False
        return True, "Stopping current monitor"

    @ocs_agent.param('channel', type=int, choices=[1, 2, 3])
    def get_voltage(self, session, params=None):
        """get_voltage(channel)

        **Task** - Measure and return the voltage of the power supply.

        Parameters:
            channel (int): Channel number (1, 2, or 3).

        Examples:
            Example for calling in a client::

                client.get_voltage(channel=1)

        Notes:
            An example of the session data::

                >>> response.session['data']
                {'timestamp': 1723671503.4899583,
                 'channel': 1,
                 'voltage': 0.0512836}

        """
        chan = params['channel']
        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                if self.psu.get_output(chan):
                    data = {
                        'timestamp': time.time(),
                        'channel': chan,
                        'voltage': self.psu.get_volt(chan)
                    }

                    session.data = data
                else:
                    return False, "Cannot measure output when output is disabled."
            else:
                return False, "Could not acquire lock"
        return True, 'Channel {} voltage measured'.format(chan)

    @ocs_agent.param('channel', type=int, choices=[1, 2, 3])
    @ocs_agent.param('volts', type=float, check=lambda x: 0 <= x <= 30)
    def set_voltage(self, session, params=None):
        """set_voltage(channel, volts)

        **Task** - Set the voltage of the power supply.

        Parameters:
            channel (int): Channel number (1, 2, or 3).
            volts (float): Voltage to set. Must be between 0 and 30.

        """

        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                self.psu.set_volt(params['channel'], params['volts'])
            else:
                return False, "Could not acquire lock"

        return True, 'Set channel {} voltage to {}'.format(params['channel'], params['volts'])

    @ocs_agent.param('channel', type=int, choices=[1, 2, 3])
    def get_current(self, session, params=None):
        """get_current(channel)

        **Task** - Measure and return the current of the power supply.

        Parameters:
            channel (int): Channel number (1, 2, or 3).

        Examples:
            Example for calling in a client::

                client.get_current(channel=1)

        Notes:
            An example of the session data::

                >>> response.session['data']
                {'timestamp': 1723671503.4899583,
                 'channel' : 1,
                 'current': 0.0103236}

        """
        chan = params['channel']
        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                # Check if output is enabled before measuring
                if self.psu.get_output(chan):
                    data = {
                        'timestamp': time.time(),
                        'channel': chan,
                        'current': self.psu.get_curr(chan)
                    }
                    session.data = data
                else:
                    return False, "Cannot measure output when output is disabled."
            else:
                return False, "Could not acquire lock"
        return True, 'Channel {} current measured'.format(chan)

    @ocs_agent.param('channel', type=int, choices=[1, 2, 3])
    @ocs_agent.param('current', type=float)
    def set_current(self, session, params=None):
        """set_current(channel, current)

        **Task** - Set the current of the power supply.

        Parameters:
            channel (int): Channel number (1, 2, or 3).
            current (float): Current to set.

        """
        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                self.psu.set_curr(params['channel'], params['current'])
            else:
                return False, "Could not acquire lock"

        return True, 'Set channel {} current to {}'.format(params['channel'], params['current'])

    @ocs_agent.param('channel', type=int, choices=[1, 2, 3])
    def get_output(self, session, params=None):
        """get_output(channel)

        **Task** - Check if channel ouput is enabled or disabled.

        Parameters:
            channel (int): Channel number (1, 2, or 3).

        Examples:
            Example for calling in a client::

                client.get_output(channel=1)

        Notes:
            An example of the session data::

                >>> response.session['data']
                {'timestamp': 1723671503.4899583,
                 'channel': 1,
                 'state': 1}

        """
        chan = params['channel']
        enabled = False
        data = {}
        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                data = {
                    'timestamp': time.time(),
                    'channel': chan,
                    'state': self.psu.get_output(chan)
                }
                session.data = data
                enabled = bool(data['state'])
            else:
                return False, "Could not acquire lock."
        if enabled:
            return True, 'Channel {} output is currently enabled.'.format(chan)
        else:
            return True, 'Channel {} output is currently disabled.'.format(chan)

    @ocs_agent.param('channel', type=int, choices=[1, 2, 3])
    @ocs_agent.param('state', type=bool)
    def set_output(self, session, params=None):
        """set_output(channel, state)

        **Task** - Turn a channel on or off.

        Parameters:
            channel (int): Channel number (1, 2, or 3).
            state (bool): True for on, False for off.

        """
        with self.lock.acquire_timeout(1) as acquired:
            if acquired:
                self.psu.set_output(params['channel'], params['state'])
            else:
                return False, "Could not acquire lock"

        return True, 'Initialized PSU.'


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address')
    pgroup.add_argument('--gpib-slot')
    pgroup.add_argument('--port')
    pgroup.add_argument('--mode', type=str, default='acq',
                        choices=['init', 'acq'])

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()
    args = site_config.parse_args(agent_class='ScpiPsuAgent',
                                  parser=parser,
                                  args=args)

    init_params = False
    if args.mode == 'acq':
        init_params = {'auto_acquire': True}
    agent, runner = ocs_agent.init_site_agent(args)

    p = ScpiPsuAgent(agent, args.ip_address, args.gpib_slot, port=args.port)

    agent.register_task('init', p.init, startup=init_params)
    agent.register_task('set_voltage', p.set_voltage)
    agent.register_task('set_current', p.set_current)
    agent.register_task('set_output', p.set_output)

    agent.register_task('get_voltage', p.get_voltage)
    agent.register_task('get_current', p.get_current)
    agent.register_task('get_output', p.get_output)

    agent.register_process('monitor_output', p.monitor_output, p.stop_monitoring)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



# For a deeper understanding of the pid command message syntax refer to: https://assets.omega.com/manuals/M3397.pdf
import socket
import time
from dataclasses import dataclass, field
from typing import Optional, Union


def retry_multiple_times(loops=3):
    def dec_wrapper(func):
        def inner(*args, **kwargs):
            for i in range(loops):
                try:
                    return func(*args, **kwargs)
                except BaseException:
                    time.sleep(0.2)
            print(f'Could not complete {func.__name__} after {loops} attempt(s)')
            return DecodedResponse(msg_type='error', msg='Read Error')
        return inner
    return dec_wrapper


@dataclass
class DecodedResponse:
    msg_type: str
    msg: str
    measure: Optional[Union[int, float]] = field(default=None)


class PID:
    """Class to communicate with the Omega CNi16D54-EIT PID controller.

    Args:
        ip (str): IP address for the controller.
        port (int): Port number for the socket connection.
        verb (bool): Verbose output setting. Defaults to False.

    Attributes:
        verb (bool): Verbose output setting.
        hex_freq (str): Currently declared rotation frequency in hexadecimal.
        conn (socket.socket): Socket object with open connection to the PID
            controller.

    """

    def __init__(self, ip, port, verb=False):
        self.verb = verb
        self.ip = ip
        self.port = port
        self.hex_freq = '00000'
        self.conn = self._establish_connection(self.ip, int(self.port))

    @staticmethod
    def _establish_connection(ip, port, timeout=2):
        """Connect to PID controller.

        Args:
            ip (str): IP address for controller.
            port (int): Port number for socket connection.
            timeout (float): Time in seconds to wait for comms until timeout
                occurs.

        Returns:
            socket.socket: Socket object with open connection to the PID
                controller.

        """
        conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        conn.settimeout(timeout)
        # unit tests might fail on first connection attempt
        attempts = 3
        for attempt in range(attempts):
            try:
                conn.connect((ip, port))
                break
            except (ConnectionRefusedError, OSError):
                print(f"Failed to connect to device at {ip}:{port}")
        else:
            raise RuntimeError('Could not connect to PID controller')
        return conn

    @staticmethod
    def _convert_to_hex(value, decimal):
        """Converts the user input into a format the PID controller can
        read.

        Args:
            value (float): value to be converted
            decimal (int): number of decimal places needed by the PID
                           controller. Depends on what the converted
                           value is needed for (e.g. the setpoint values
                           need take the form X.XXX so decimal=3)

        Returns:
            str: encoded hex string

        Examples:
            _convert_to_hex(0, 3) -> '0000'
            _convert_to_hex(2.0, 3) -> '07D0'

        """
        temp_value = hex(int(10**decimal * float(value)))
        return ('0000' + str(temp_value)[2:].upper())[-4:]

    @staticmethod
    def _get_scale_hex(num, corr):
        """Gets the exponent in scientific notation."""
        expo = int(str(num * 10**40 + 0.01).split('+')[1]) - 40
        digits = round(num * 10**(-expo + 4))

        expo_hex = str(hex(corr - expo))[2:]
        digits_hex = ('00000' + str(hex(digits))[2:])[-5:]

        return expo_hex + digits_hex

    ######################################################################
    # Main Processes
    ######################################################################

    def set_direction(self, direction):
        """Sets the direction if the CHWP.

        Modifies the ``direction`` attribute.

        Command messages:
            W024XXXXX:
            - W: write type command
            - 02: target pid 2 setpoint
            - 4: decimal point format X.XXX
            - XXXXX: write value

        Args:
            direction (int): 0 for forward and 1 for backwards

        """
        if direction == '0':
            if self.verb:
                print('Forward')
            resp = self.send_message("*W02400000")
        elif direction == '1':
            if self.verb:
                print('Reverse')
            resp = self.send_message("*W02401388")

        if self.verb:
            print(self.return_messages([resp])[0])

    def declare_freq(self, freq):
        """Declare to memory what the CHWP frequency should be.

        **Note:** Does not actually change the frequency.

        Args:
            freq (float): Set freqency in Hz. Must be less than 3.5.

        """
        if float(freq) <= 3.5:
            self.hex_freq = '0' + self._convert_to_hex(freq, 3)
            if self.verb:
                print('Frequency Setpoint = ' + str(freq) + ' Hz')
        else:
            if self.verb:
                print('Invalid Frequency')

    def tune_stop(self):
        """Set the setpoint to 0 Hz and stop the CHWP.

        Command messages:
            W0C83:
            - W: write type command
            - 0C: target pid 1 action type
            - 83: set action type to direct
            W01400000
            - W: write type command
            - 01: target pid 1 setpoint
            - 4: decimal point format X.XXX
            - 00000: write value
            R01:
            - R: read type command
            - 01: target pid 1 setpoint
            Z02:
            - Z: reset type command
            - 02: target entire controller

        """
        if self.verb:
            print('Starting Stop')

        responses = []
        responses.append(self.send_message("*W0C83"))
        responses.append(self.send_message("*W01400000"))
        responses.append(self.send_message("*R01"))
        responses.append(self.send_message("*Z02"))
        if self.verb:
            print(responses)
            print(self.return_messages(responses))

        stop_params = [0.2, 0, 0]
        self.set_pid(stop_params)

    def tune_freq(self):
        """Set the setpoint to currently declared frequency.

        To declare the frequency, use ``PID.declare_freq()``.

        Command messages:
            W0C81:
            - W: write type command
            - 0C: target pid 1 action type
            - 81: set action type to reverse
            W014XXXXX
            - W: write type command
            - 01: target pid 1 setpoint
            - 4: decimal point format X.XXX
            - XXXXX: write value
            R01:
            - R: read type command
            - 01: target pid 1 setpoint
            Z02:
            - Z: reset type command
            - 02: target entire controller

        """
        if self.verb:
            print('Starting Tune')

        responses = []
        responses.append(self.send_message("*W0C81"))
        responses.append(self.send_message(f"*W014{self.hex_freq}"))
        responses.append(self.send_message("*R01"))
        responses.append(self.send_message("*Z02"))
        if self.verb:
            print(responses)
            print(self.return_messages(responses))

        tune_params = [0.2, 63, 0]
        self.set_pid(tune_params)

    @retry_multiple_times(loops=3)
    def get_freq(self):
        """Returns the current frequency of the CHWP.

        Command messages:
            X01:
            - X: read (decimal) type command
            - 01: target pid 1 value

        """
        if self.verb:
            print('Finding CHWP Frequency')

        responses = []
        responses.append(self.send_message("*X01"))
        decoded_resp = self.return_messages(responses)[0]
        if self.verb:
            print(responses)
            print(decoded_resp)
        if decoded_resp.msg_type == 'measure':
            return decoded_resp
        elif decoded_resp.msg_type == 'error':
            print(f"Error reading freq: {decoded_resp.msg}")
            raise ValueError
        else:
            print("Unknown freq response")
            raise ValueError

    @retry_multiple_times(loops=3)
    def get_target(self):
        """Returns the target frequency of the CHWP.

        Command messages:
            R01:
            - R: read type command
            - 01: target pid 1 setpoint

        """
        if self.verb:
            print('Finding target CHWP Frequency')

        responses = []
        responses.append(self.send_message("*R01"))
        decoded_resp = self.return_messages(responses)[0]
        if self.verb:
            print(responses)
            print(decoded_resp)
        if decoded_resp.msg_type == 'read':
            return decoded_resp
        elif decoded_resp.msg_type == 'error':
            print(f"Error reading target: {decoded_resp.msg}")
            raise ValueError
        else:
            print('Unknown target response')
            raise ValueError

    @retry_multiple_times(loops=3)
    def get_direction(self):
        """Get the current rotation direction.

        Returns:
            int: 0 for forward and 1 for backwards

        Command messages:
            R02:
            - R: read type command
            - 02: target pid 2 setpoint

        """
        if self.verb:
            print('Finding CHWP Direction')

        responses = []
        responses.append(self.send_message("*R02"))
        decoded_resp = self.return_messages(responses)[0]
        if self.verb:
            print(responses)
            print(decoded_resp)
        if decoded_resp.msg_type == 'read':
            return decoded_resp
        elif decoded_resp.msg_type == 'error':
            print(f"Error reading direction: {decoded_resp.msg}")
            raise ValueError
        else:
            print('Unknown direction response')
            raise ValueError

    def set_pid(self, params):
        """Sets the PID parameters of the controller.

        Command messages:
            W17XXXX
            - W: write type command
            - 17: target pid 1 p param
            - XXXX: write value
            W18XXXX
            - W: write type command
            - 18: target pid 1 i param
            - XXXX: write value
            W19XXXX
            - W: write type command
            - 19: target pid 1 d param
            - XXXX: write value
            Z02:
            - Z: reset type command
            - 02: target entire controller

        """
        if self.verb:
            print('Setting PID Params')

        p_value = self._convert_to_hex(params[0], 3)
        i_value = self._convert_to_hex(params[1], 0)
        d_value = self._convert_to_hex(params[2], 1)

        responses = []
        responses.append(self.send_message(f"*W17{p_value}"))
        responses.append(self.send_message(f"*W18{i_value}"))
        responses.append(self.send_message(f"*W19{d_value}"))
        responses.append(self.send_message("*Z02"))
        time.sleep(2)
        if self.verb:
            print(responses)
            print(self.return_messages(responses))

    def set_scale(self, slope, offset):
        """Set the conversion between feedback voltage and approximate
        frequency.

        Command messages:
            W14XXXXX:
            - W: write type command
            - 14: target feedback scale
            - XXXXX: write value
            W03XXXXX:
            - W: write type command
            - 03: target feedback offset
            - XXXXX: write value
            Z02:
            - Z: reset type command
            - 02: target entire controller

        """
        slope_hex = self._get_scale_hex(slope, 1)
        offset_hex = self._get_scale_hex(offset, 2)

        responses = []
        responses.append(self.send_message(f"*W14{slope_hex}"))
        responses.append(self.send_message(f"*W03{offset_hex}"))
        responses.append(self.send_message("*Z02"))
        if self.verb:
            print(responses)
            print(self.return_messages(responses))

    ######################################################################
    # Messaging
    ######################################################################

    def send_message(self, msg):
        """Send message over TCP to the PID controller.

        Args:
            msg (str): Command to send to the controller.

        Returns:
            str: Response from the controller.

        """
        for attempt in range(2):
            try:
                self.conn.sendall((msg + '\r\n').encode())
                time.sleep(0.5)  # Don't send messages too quickly
                data = self.conn.recv(4096).decode().strip()
                return data
            except (socket.timeout, OSError):
                print("Caught timeout waiting for response from PID controller. "
                      + "Trying again...")
                time.sleep(1)
                if attempt == 1:
                    print("Resetting connection")
                    self.conn.close()
                    self.conn = self._establish_connection(self.ip, int(self.port))
                    return self.send_message(msg)

    def return_messages(self, msg):
        """Decode list of responses from PID controller and return useful
        values.

        Args:
            msg (list): List of messages to decode.

        Returns:
            list: DecodedResponse

        """
        return self._decode_array(msg)

    @staticmethod
    def _decode_array(input_array):
        """Helper function to parse the individual response strings.

        Each function calls a series of commands which each have a reponse string.
        The strings are arranged into an array and sent here to be decoded. For each
        individual string, the first character is the action type:
            - R: read (hex)
            - W: write (hex)
            - E: enable
            - D: disable
            - P: put
            - G: get
            - X: read (decimal)
            - Z: reset
        Following the action type, the next two characters are the command type. The
        supported command type depends on the action type:
            - R01: read setpoint for pid 1 (rotation frequency setpoint)
            - R02: read setpoint for pid 2 (rotation direction setpoint)
            - W01: write setpoint for pid 1 (rotation frequency setpoint)
            - W02: write setpoint for pid 2 (rotation direction setpoint)
            - W0C: write action type for pid 1 (how to interpret sign of (setpoint-value))
            - X01: read value for pid 1 (current rotation frequency)
        "?" character indicates the error messages.
        The helper function goes through the raw response strings and replaces them
        with their decoded values.

        Args:
            input_array (list): List of str messages to decode

        Returns:
            list: DecodedResponse

        """
        output_array = list(input_array)

        for index, string in enumerate(list(input_array)):
            if not isinstance(string, str):
                output_array[index] = DecodedResponse(msg_type='error', msg='Unrecognized response')
                continue
            header = string[0]
            if '?' in string:
                output_array[index] = PID._decode_error(string)
            elif header == 'R':
                output_array[index] = PID._decode_read(string)
            elif header == 'W':
                output_array[index] = PID._decode_write(string)
            elif header == 'E':
                output_array[index] = DecodedResponse(msg_type='enable', msg='PID Enabled')
            elif header == 'D':
                output_array[index] = DecodedResponse(msg_type='disable', msg='PID Disabled')
            elif header == 'P':
                pass
            elif header == 'G':
                pass
            elif header == 'X':
                output_array[index] = PID._decode_measure(string)
            elif header == 'Z':
                output_array[index] = DecodedResponse(msg_type='reset', msg='PID Reset')
            else:
                output_array[index] = DecodedResponse(msg_type='error', msg='Unrecognized response')

        return output_array

    @staticmethod
    def _decode_error(string):
        """Helper function to decode error messages

        Args:
            string (str): Error message type string to decode

        Returns:
            DecodedResponse

        """
        if '?+9999.' in string:
            return DecodedResponse(msg_type='error', msg='Exceed Maximum Error')
        elif '?43' in string:
            return DecodedResponse(msg_type='error', msg='Command Error')
        elif '?46' in string:
            return DecodedResponse(msg_type='error', msg='Format Error')
        elif '?50' in string:
            return DecodedResponse(msg_type='error', msg='Parity Error')
        elif '?56' in string:
            return DecodedResponse(msg_type='error', msg='Serial Device Address Error')
        else:
            return DecodedResponse(msg_type='error', msg='Unrecognized Error')

    @staticmethod
    def _decode_read(string):
        """Helper function to decode "read (hex)" type response strings

        Specific decoding procedure depends on response string type:
            - R01 (pid 1 setpoint): convert hex value into decimal (X.XXX format).
                                    Returns decimal value
            - R02 (pid 2 setpoint): convert hex value into decimal (X.XXX format)
                                    and compair to mean value to determine rotation
                                    direction. Returns rotation direction

        Examples:
            _decode_read('R01400000') -> 0.0
            _decode_read('R014007D0') -> 2.0
            _decode_read('R02400000') -> 0
            _decode_read('R024007D0') -> 1

        Args:
            string (str): Read (hex) type string to decode

        Returns:
            DecodedResponse

        """
        end_string = string.split('\r')[-1]
        read_type = end_string[1:3]
        if len(end_string) != 9:
            return DecodedResponse(msg_type='error', msg='Unrecognized Read Length')
        # Decode target
        if read_type == '01':
            target = float(int(end_string[4:], 16) / 1000.)
            return DecodedResponse(msg_type='read', msg='Setpoint = ' + str(target), measure=target)
        # Decode direction
        elif read_type == '02':
            if int(end_string[4:], 16) / 1000. > 2.5:
                return DecodedResponse(msg_type='read', msg='Direction = Reverse', measure=1)
            else:
                return DecodedResponse(msg_type='read', msg='Direction = Forward', measure=0)
        else:
            return DecodedResponse(msg_type='error', msg='Unrecognized Read Type')

    @staticmethod
    def _decode_write(string):
        """Helper function to decode "write (hex)" type response strings

        Args:
            string (str): Write (hex) type string to decode

        Returns:
            DecodedResponse

        """
        write_type = string[1:]
        if write_type == '01':
            return DecodedResponse(msg_type='write', msg='Changed Setpoint')
        elif write_type == '02':
            return DecodedResponse(msg_type='write', msg='Changed Direction')
        elif write_type == '0C':
            return DecodedResponse(msg_type='write', msg='Changed Action Type')
        elif write_type == '17':
            return DecodedResponse(msg_type='write', msg='Changed PID 1 P Param')
        elif write_type == '18':
            return DecodedResponse(msg_type='write', msg='Changed PID 1 I Param')
        elif write_type == '19':
            return DecodedResponse(msg_type='write', msg='Changed PID 1 D Param')
        else:
            return DecodedResponse(msg_type='error', msg='Unrecognized Write')

    @staticmethod
    def _decode_measure(string):
        """Helper function to decode "read (decimal)" type response strings

        Deconding is done in the following way
            - X01 (pid 1 value): removes header and returns decimal value

        Args:
            string (str): Read (decimal) type string to decode

        Return:
            DecodedReponse
        """
        end_string = string.split('\r')[-1]
        measure_type = end_string[1:3]
        if measure_type == '01':
            freq = float(end_string[3:])
            return DecodedResponse(msg_type='measure', msg='Current frequency = ' + str(freq), measure=freq)
        else:
            return DecodedResponse(msg_type='error', msg='Unrecognized Measure')



#!/usr/bin/env python3
"""Module to read stimulator encoder.
"""
import fcntl
import mmap
import os
import sys
from pathlib import Path
from queue import Queue
from threading import Thread
from time import sleep, time

import numpy as np

ADDR_AXI = 0x80020000
PATH_DEV_BASE = Path(f'/sys/devices/platform/axi/{ADDR_AXI:08x}.str_rd/uio')
PATH_LOCK = Path('/tmp/').joinpath('.stim-lock')


class StimEncError(Exception):
    """
    Exception rased by stimulator encoder reader.
    """


class StimEncTime:
    """
    Stimulator encoder time.

    Parameter
    ---------
    time_raw : int
        Raw time format from TSU.
        [sec 48 bits][nsec 30 bits][sub-nsec 16 bits]
    """

    def __init__(self, time_raw: int):
        self._time_raw = time_raw

    @property
    def sec(self) -> int:
        """
        Seconds part of timestamp.

        Returns
        -------
        sec : int
            Seconds part of timestamp.
        """
        return self._time_raw >> 46

    @property
    def nsec(self) -> int:
        """
        Nano second part of timestamp.

        Returns
        -------
        nsec : int
            Nano-sec part of timestamp.
        """
        return (0x_00000000_00003fff_ffff0000 & self._time_raw) >> 16

    @property
    def tai(self) -> float:
        """
        Time in seconds from TAI epoch.

        Returns
        -------
        tai : float
            Seconds from TAI epoch.
        """
        return self.sec + (self.nsec / 1e9)


class StimEncData:
    """
    Stimulator encoder data.

    Parameter
    ---------
    data_raw : ndarray
        Raw data from PL FIFO.
    """

    def __init__(self, data_bytes):
        self._data_bytes = data_bytes
        self._data_int = int(data_bytes[0]) + (int(data_bytes[1]) << 32) + (int(data_bytes[2]) << 64)
        self._utime = time()

    @property
    def state(self) -> int:
        return (self._data_int & 0xC0_00_00_00_00000000_00000000) >> 94

    @property
    def time_raw(self) -> int:
        """94 bit TSU timestamp.

        Returns
        -------
        time_raw : int
            94 bit TSU timestamp.
        """
        return self._data_int & 0x3F_FF_FF_FF_FFFFFFFF_FFFFFFFF

    @property
    def time(self) -> StimEncTime:
        """
        TSU timestamp.

        Returns
        -------
        time : StimTime
            TSU timestamp abstraction.
        """
        return StimEncTime(self.time_raw)

    @property
    def utime(self) -> float:
        """
        Unix timestamp at class creation.

        Returns
        -------
        utime : Unix timestamp when this object is created.
        """
        return self._utime

    def __str__(self):
        return f'time={int(self.time.g3) / 1e8:.8f} data={self.state:02b}'


def get_path_dev() -> Path:
    """
    Acquire devicefile path for `str_rd` IP core.

    Returns
    -------
    path_dev : Path
        Path to the device file.
    """
    if not PATH_DEV_BASE.exists():
        raise StimEncError('Device is not found. Check firmware and device tree.')

    name_dev = list(PATH_DEV_BASE.glob('uio*'))[0].name
    path_dev = Path(f'/dev/{name_dev}')

    return path_dev


class StimEncReader:
    """Class to read encoder data.

    Parameters
    ----------
    path_dev : str or pathlib.Path
        Path to the generic-uio device file for str_rd IP.
    path_lock : str or pathlib.Path
        Path to the lockfile.
    """

    def __init__(self, path_dev, path_lock=PATH_LOCK, verbose=True):
        # Verbose level
        self._verbose = verbose

        # Locking
        self._fp_lock = open(path_lock, 'w')
        try:
            fcntl.flock(self._fp_lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            raise StimEncError('locked.')

        # Connection
        self._path_dev = path_dev
        self._dfile = os.open(self._path_dev, os.O_RDONLY | os.O_SYNC)
        self._dev = mmap.mmap(self._dfile, 0x100, mmap.MAP_SHARED, mmap.PROT_READ, offset=0)

        # Data FIFO
        self.fifo: "Queue[StimEncData]" = Queue()

        # Runner
        self._thread = None
        self._running = False

    def __del__(self):
        if self._running:
            self.stop()
        fcntl.flock(self._fp_lock, fcntl.LOCK_UN)
        self._dev.close()
        os.close(self._dfile)

        self._eprint('Fin.')

    def _eprint(self, errmsg):
        if self._verbose:
            sys.stderr.write(f'{errmsg}\r\n')

    def _get_info(self):
        data = np.frombuffer(self._dev, np.uint32, 4, offset=0)
        r_len = data[0]
        w_len = data[1]
        residue = data[2]

        return r_len, w_len, residue

    def _get_data(self) -> StimEncData:
        data = np.frombuffer(self._dev, np.uint32, 4, offset=16)

        return StimEncData(data)

    def fill(self):
        """
        Get data from PL fifo and put into software fifo.
        """
        while True:
            r_len, w_len, residue = self._get_info()

            if (r_len == 0) and (residue == 0):
                break

            self.fifo.put(self._get_data())

    def _loop(self):
        while self._running:
            self.fill()
            sleep(0.1)

    def run(self):
        """
        Run infinite loop of data filling.
        """
        self._running = True
        self._thread = Thread(target=self._loop)
        self._thread.start()

    def stop(self):
        if not self._running:
            raise StimEncError('Not started yet.')

        self._running = False
        self._thread.join()


def main():
    """Main function to boot infinite loop"""
    stim_enc = StimEncReader(get_path_dev(), verbose=True)

    # Filler loop
    fd = open('test.dat', 'w')
    stim_enc.run()
    while True:
        try:
            while not stim_enc.fifo.empty():
                data = stim_enc.fifo.get()
                print(data)
                fd.write(str(data) + '\n')
            sleep(0.1)
        except KeyboardInterrupt:
            stim_enc.stop()
            fd.close()
            break

    print('Fin.')


if __name__ == '__main__':
    main()



# Copyright (C) 2023-2024 Simons Observatory Collaboration
# See top-level LICENSE.txt file for more information.
"""Agent to capture images from cameras which support the RTSP protocol.
"""
import os
import time
from datetime import datetime, timedelta, timezone

import cv2
import ocs
import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock

txaio.use_twisted()

from socs.common.camera import (CircularMediaBuffer, FakeCamera,
                                MotionDetector, image_read_callback,
                                image_write_callback, video_write_callback)


class RTSPCameraAgent:
    """Agent to support image capture from RTSP cameras.

    This Agent captures images and writes them to a feed, as well as saving frames
    to disk.  The camera can also be triggered to record....

    Args:
        agent (OCSAgent): OCSAgent object from :func:`ocs.ocs_agent.init_site_agent`.
        directory (str): The path to image storage for this camera.
        address (str): The hostname or IP address of the camera.
        user (str): The user name for camera access.
        password (str): The password for camera access.
        seconds (int): The seconds between snapshots.
        port (int): The RTSP port to use (default is standard 554).
        urlpath (str): The remaining URL for the camera, which might include
            options like channel and subtype.  This will depend on the manufacturer.
        jpeg_quality (int): The JPEG quality for snapshot images (0-100).
        max_snapshot_files (int): The maximum number of snapshots to keep.
        record_fps (float): The frames per second for recorded video.
        record_duration (int): The number of seconds for each recorded video.
        max_record_files (int): The maximum number of recordings to keep.
        motion_start (str): ISO time (HH:MM:SS+-zz:zz) to start motion detection.
        motion_stop (str): ISO time (HH:MM:SS+-zz:zz) to stop motion detection.
        disable_motion (bool): If True, disable motion detection.
        fake (bool): If True, ignore camera settings and generate fake video
            for testing.

    Attributes:
        agent (OCSAgent): OCSAgent object from :func:`ocs.ocs_agent.init_site_agent`.
        log (txaio.tx.Logger): Logger object used to log events within the
            Agent.
        lock (TimeoutLock): TimeoutLock object used to prevent simultaneous
            commands being sent to hardware.

    """

    def __init__(
        self,
        agent,
        directory,
        address,
        user,
        password,
        seconds=10,
        port=554,
        urlpath=None,
        jpeg_quality=20,
        max_snapshot_files=10000,
        record_fps=20.0,
        record_duration=60,
        max_record_files=100,
        motion_start=None,
        motion_stop=None,
        disable_motion=False,
        fake=False,
    ):
        self.agent = agent
        self.topdir = directory
        self.log = agent.log
        self.lock = TimeoutLock(default_timeout=5)

        self.address = address
        self.port = port
        self.user = user
        self.password = password
        self.seconds = seconds
        self.urlpath = urlpath
        self.fake = fake
        self.motion_start = motion_start
        self.motion_stop = motion_stop
        self.motion_detect = not disable_motion

        if self.urlpath is None:
            # Try the string for the Dahua cameras at the site
            self.urlpath = "/cam/realmonitor?channel=1&subtype=0"

        self.connection = f"rtsp://{self.user}:{self.password}"
        self.connection += f"@{self.address}:{self.port}{self.urlpath}"

        # We will store recordings and snapshots to separate subdirs
        if not os.path.isdir(self.topdir):
            os.makedirs(self.topdir)
        self.img_dir = os.path.join(self.topdir, "snapshots")
        self.vid_dir = os.path.join(self.topdir, "recordings")
        self.record_duration = record_duration
        self.record_fps = record_fps

        # Create the image buffer on disk
        self.img_buffer = CircularMediaBuffer(
            self.img_dir,
            "img",
            "jpg",
            max_snapshot_files,
            image_write_callback,
            read_callback=image_read_callback,
            recent=3,
            quality=jpeg_quality,
        )

        # Create the recording buffer on disk
        self.vid_buffer = CircularMediaBuffer(
            self.vid_dir,
            "vid",
            "mp4",
            max_record_files,
            video_write_callback,
            read_callback=None,
            recent=0,
            fps=record_fps,
        )

        # Register OCS feed
        self.feed_name = f"RTSP_{self.address}"

        agg_params = {"frame_length": self.seconds}  # [sec]
        self.agent.register_feed(
            self.feed_name,
            record=True,
            agg_params=agg_params,
            buffer_time=1.0,
        )

    def _in_motion_time_range(self):
        """Determine if we are in the valid time range for motion detection."""
        if self.motion_start is None or self.motion_stop is None:
            # We are not using the start / stop time range, so all times are valid
            return True

        # The current time in UTC
        curtime = datetime.now(tz=timezone.utc)

        # Convert the start / stop times to datetimes based on today
        curdaystr = f"{curtime.year}-{curtime.month:02d}-{curtime.day:02d}"

        # The datetimes for start/stop today
        def _dt_convert(timestr):
            tstr = f"{curdaystr}T{timestr}"
            try:
                tm = datetime.strptime(tstr, "%Y-%m-%dT%H:%M:%S%z")
            except ValueError:
                tm = datetime.strptime(tstr, "%Y-%m-%dT%H:%M:%S")
                msg = f"Motion time '{timestr}' is not "
                msg += "timezone-aware.  Assuming UTC."
                self.log.warning(msg)
                tm = tm.replace(tzinfo=timezone.utc)
            return tm
        start = _dt_convert(self.motion_start)
        stop = _dt_convert(self.motion_stop)

        if stop <= start:
            # We are starting today and stopping tomorrow
            stop += timedelta(days=1)

        if curtime > start and curtime < stop:
            return True
        else:
            return False

    def _init_stream(self):
        """Connect to camera and initialize video stream."""
        if self.fake:
            cap = FakeCamera()
        else:
            cap = cv2.VideoCapture(self.connection)
        if not cap.isOpened():
            self.log.error(f"Cannot open RTSP stream at {self.connection}")
            return None

        return cap

    @ocs_agent.param("test_mode", default=False, type=bool)
    def acq(self, session, params=None):
        """acq(test_mode=False)

        **Process** - Capture frames from an RTSP camera.

        Args:
            test_mode (bool):  Run the Process loop only once. Meant only
                for testing.  Default is False.

        Notes:
            Individual frames are written to a circular disk buffer.  Metadata
            about the captured images is stored in the session data.  The format
            of this is::

                >>> response.session['data']
                {
                    'address': 'camera-c1.example.org',
                    'timestamp': 1701983575.123456,
                    'path': '/ocs/cameras_rtsp/c1/img_2023-12-29T02:44:47+00:00.jpg',
                }

        """
        pm = Pacemaker(1 / self.seconds, quantize=False)
        pmgrab = Pacemaker(self.record_fps, quantize=True)

        frames_per_snapshot = int(self.seconds * self.record_fps)

        self.is_streaming = True

        # Open camera stream
        cap = self._init_stream()
        connected = True
        if not cap:
            connected = False

        # Tracking state of whether we are currently recording motion detection
        detecting = False
        detect_start = None
        record_frames = int(self.record_fps * self.record_duration)
        motion_detector = MotionDetector()

        snap_count = 0
        while self.is_streaming:
            if not connected:
                self.log.info("Trying to reconnect.")
                cap = self._init_stream()
                if not cap:
                    pm.sleep()
                    continue

            # Use UTC
            timestamp = time.time()
            data = dict()

            for iframe in range(frames_per_snapshot):
                pmgrab.sleep()
                # Grab an image
                _ = cap.grab()
            success, image = cap.retrieve()
            if not success:
                msg = "Failed to retrieve snapshot image from stream"
                self.log.error(msg)
                connected = False
                continue
            connected = True

            # Motion detection.  We ignore the first few snapshots and also
            # any changes that happen while we are already recording.
            if snap_count < 4:
                skip = True
            else:
                skip = False
            if detecting:
                if (snap_count - detect_start) * frames_per_snapshot > record_frames:
                    # We must have finished recording
                    detecting = False
                    skip = False
                else:
                    # We are still recording
                    skip = True
            if self.motion_detect and self._in_motion_time_range():
                image, movement = motion_detector.process(image, skip=skip)
                if movement:
                    # Start recording
                    detecting = True
                    detect_start = snap_count
                    rec_stat, rec_msg, _ = self.agent.start(
                        "record", params={"test_mode": False}
                    )
                    if rec_stat != ocs.OK:
                        self.log.error(f"Problem with motion capture: {rec_msg}")

            # Save to circular buffer
            self.img_buffer.store(image)

            # Get the saved path
            path = self.img_buffer.fetch_index(-1)[0]

            # Fill data
            data = {
                "address": self.address,
                "timestamp": timestamp,
                "path": path,
                "connected": connected
            }

            # Update session.data and publish
            session.data = data
            self.log.debug(f"{data}")

            message = {
                "block_name": "cameras",
                "timestamp": timestamp,
                "data": {
                    "address": self.address,
                    "path": path,
                    "connected": connected
                },
            }
            session.app.publish_to_feed(self.feed_name, message)

            if params["test_mode"]:
                break
            snap_count += 1
            pm.sleep()

        # Flush buffer and stop the data stream
        self.agent.feeds[self.feed_name].flush_buffer()

        # Release stream
        cap.release()
        return True, "Acquisition finished"

    def _stop_acq(self, session, params=None):
        """_stop_acq()
        **Task** - Stop task associated with acq process.
        """
        if self.is_streaming:
            self.is_streaming = False
            return True, "Stopping Acquisition"
        else:
            return False, "Acq is not currently running"

    def record(self, session, params=None):
        """Record video stream.

        **Task** - Record video for fixed timespan.

        Parameters:
            None

        """
        session.set_status("running")

        with self.lock.acquire_timeout(0, job="record") as acquired:
            if not acquired:
                self.log.warn(
                    "Could not start recording because "
                    "{} is already running".format(self.lock.job)
                )
                return False, "Only one simultaneous recording per camera allowed"

            pm = Pacemaker(self.record_fps, quantize=True)

            # Open camera stream
            self.log.info("Recording:  opening camera stream")
            cap = self._init_stream()
            if not cap:
                return False, "Cannot connect to camera."

            # Total number of frames
            total_frames = int(self.record_fps * self.record_duration)

            msg = f"Recording:  starting {total_frames} frames "
            msg += f"({self.record_duration}s at {self.record_fps}fps)"
            self.log.info(msg)

            frames = list()
            for iframe in range(total_frames):
                if session.status != "running":
                    return False, "Aborted recording"
                pm.sleep()
                # Grab an image
                success, image = cap.read()
                if not success:
                    msg = f"Recording:  broken stream at frame {iframe}, ending"
                    self.log.error(msg)
                    break
                frames.append(image)

            # Save to circular buffer
            self.vid_buffer.store(frames)

            # Get the saved path
            path = self.vid_buffer.fetch_index(-1)[0]
            self.log.info(f"Recording:  finished {path}")

            # Cleanup
            cap.release()

        return True, "Recording finished."

    def _abort_record(self, session, params):
        if session.status == "running":
            session.set_status("stopping")


def add_agent_args(parser=None):
    """Create or append agent arguments.

    Args:
        parser (ArgumentParser): Optional input parser to use.  If None,
            a new parser will be created.

    Returns:
        (ArgumentParser): The created or modified parser.

    """
    if parser is None:
        from argparse import ArgumentParser as A

        parser = A()
    pgroup = parser.add_argument_group("Agent Options")

    pgroup.add_argument(
        "--mode",
        type=str,
        default="acq",
        choices=["acq", "test"],
        help="Starting action for the Agent.",
    )

    pgroup.add_argument(
        "--directory",
        type=str,
        required=True,
        help="Directory for media buffers (snapshots and recordings)",
    )

    pgroup.add_argument(
        "--address",
        type=str,
        required=True,
        help="Hostname or IP address of camera",
    )

    pgroup.add_argument(
        "--user",
        type=str,
        required=True,
        help="User name for camera access",
    )

    pgroup.add_argument(
        "--password",
        type=str,
        required=True,
        help="Password for camera access",
    )

    pgroup.add_argument(
        "--motion_start",
        type=str,
        default=None,
        required=False,
        help="ISO 8601 time (HH:MM:SS+-zz:zz) to begin motion detection",
    )

    pgroup.add_argument(
        "--motion_stop",
        type=str,
        default=None,
        required=False,
        help="ISO 8601 time (HH:MM:SS+-zz:zz) to end motion detection",
    )

    pgroup.add_argument(
        "--snapshot_seconds",
        type=int,
        required=False,
        default=10,
        help="Number of seconds between snapshots for motion detection",
    )

    pgroup.add_argument(
        "--port",
        type=int,
        required=False,
        default=554,
        help="The RTSP port number",
    )

    pgroup.add_argument(
        "--urlpath",
        type=str,
        default=None,
        required=False,
        help="The path portion of the URL.  Default will use values for Dahua cameras.",
    )

    pgroup.add_argument(
        "--jpeg_quality",
        type=int,
        required=False,
        default=20,
        help="The JPEG quality (0-100)",
    )

    pgroup.add_argument(
        "--max_snapshot_files",
        type=int,
        required=False,
        default=17280,  # 2 days at 10s per snapshot
        help="Maximum number of images to keep in the circular buffer",
    )

    pgroup.add_argument(
        "--record_fps",
        type=float,
        required=False,
        default=20.0,
        help="The frames per second for video recording",
    )

    pgroup.add_argument(
        "--record_duration",
        type=float,
        required=False,
        default=60,
        help="The length in seconds to record video.",
    )

    pgroup.add_argument(
        "--max_record_files",
        type=int,
        required=False,
        default=120,  # Most recent 2 hours of motion capture
        help="Maximum number of images to keep in the circular buffer",
    )

    pgroup.add_argument(
        "--fake",
        action="store_true",
        required=False,
        default=False,
        help="Use an internal fake camera for acquisition",
    )

    pgroup.add_argument(
        "--disable_motion",
        action="store_true",
        required=False,
        default=False,
        help="Disable motion detection",
    )

    return parser


def main(args=None):
    txaio.make_logger()

    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = add_agent_args()
    args = site_config.parse_args(
        agent_class="RTSPCameraAgent", parser=parser, args=args
    )

    if args.mode == "acq":
        init_params = {"test_mode": False}
    elif args.mode == "test":
        init_params = {"test_mode": True}

    agent, runner = ocs_agent.init_site_agent(args)

    cam = RTSPCameraAgent(
        agent,
        args.directory,
        args.address,
        args.user,
        args.password,
        seconds=args.snapshot_seconds,
        port=args.port,
        urlpath=args.urlpath,
        jpeg_quality=args.jpeg_quality,
        max_snapshot_files=args.max_snapshot_files,
        record_fps=args.record_fps,
        record_duration=args.record_duration,
        max_record_files=args.max_record_files,
        fake=args.fake,
        motion_start=args.motion_start,
        motion_stop=args.motion_stop,
        disable_motion=args.disable_motion,
    )
    agent.register_process("acq", cam.acq, cam._stop_acq, startup=init_params)
    agent.register_task(
        "record", cam.record, aborter=cam._abort_record
    )
    runner.run(agent, auto_reconnect=True)


if __name__ == "__main__":
    main()



# Copyright (C) 2023-2024 Simons Observatory Collaboration
# See top-level LICENSE.txt file for more information.
"""Tools for working with image / video cameras.
"""
import glob
import os
import re
import shutil
from collections import deque
from datetime import datetime, timezone

import numpy as np

try:
    import cv2
    import imutils
    have_cv2 = True
except ImportError:
    have_cv2 = False


def image_read_callback(path):
    """Read callback for OpenCV images.

    This is for use with the CircularMediaBuffer class.

    Args:
        path (str):  Path to the image file.

    Returns:
        (array):  Image data.

    """
    return cv2.imread(path)


def image_write_callback(data, path, quality=20):
    """Write callback for OpenCV images.

    This is for use with the CircularMediaBuffer class.

    Args:
        data (array):  Image data.
        path (str):  Path to the image file.
        quality (int):  JPEG quality (0-100).

    Returns:
        None

    """
    cv2.imwrite(path, data, [int(cv2.IMWRITE_JPEG_QUALITY), quality])


def video_read_callback(path):
    """Read callback for OpenCV video.

    This is for use with the CircularMediaBuffer class.

    Args:
        path (str):  Path to the video file.

    Returns:
        (array):  video data.

    """
    data = list()
    cap = cv2.VideoCapture(path)
    success = True
    while success:
        success, frame = cap.read()
        if success:
            data.append(frame)
    cap.release()
    return data


def video_write_callback(data, path, fps=20.0):
    """Write callback for OpenCV video.

    This is for use with the CircularMediaBuffer class.

    Args:
        data (list):  List of frames.
        path (str):  Path to the video file.

    Returns:
        None

    """
    first = data[0]
    height = first.shape[0]
    width = first.shape[1]
    fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    out = cv2.VideoWriter(path, fourcc, fps, (width, height))
    for frame in data:
        out.write(frame)
    out.release()


class CircularMediaBuffer:
    """Class to manage a circular media file buffer on disk.

    Rather than try to guess the average file size and target disk usage,
    this class just manages a fixed number of files within a directory.

    The filenames and file data are stored as tuples in a deque.  Only the
    data for recent files is kept in memory.

    If the read callback function is None, no recent data is kept in memory.
    Any additional kwargs to the constructor are passed to the write
    callback function.

    Args:
        directory (str): The path to the directory of image files.  This
            should be a directory that is completely managed by this class
            which contains no other files.
        root_name (str): The root of each file name (e.g. "img", "vid").
        suffix (str): The file suffix (e.g. "jpg", "mp4").
        max_files (int): The maximum number of files to keep.
        write_callback (function): The function to use for writing files.
        read_callback (function): The function to use for loading recent
            files.
        recent (int): The number of recent images to keep in memory.

    """

    def __init__(
        self,
        directory,
        root_name,
        suffix,
        max_files,
        write_callback,
        read_callback=None,
        recent=3,
        **kwargs,
    ):
        self.dir = os.path.abspath(directory)
        if not os.path.isdir(self.dir):
            os.makedirs(self.dir)
        self.max_files = max_files
        self.recent = recent
        self.root = root_name
        self.suffix = suffix
        self.writer = write_callback
        self.reader = read_callback
        self.writer_opts = kwargs

        # Load current files into our store
        existing_files = filter(
            os.path.isfile,
            glob.glob(os.path.join(self.dir, f"{self.root}_*.{self.suffix}")),
        )

        # We could sort by modification time, but if files were copied that could
        # be unreliable.  Instead we sort by ISO date/time encoded in the filename.
        self._deque = deque()
        for file in sorted(existing_files):
            self._deque.append((file, None))

        # Load recent
        if self.reader is not None:
            to_load = min(len(self._deque), self.recent)
            for idx in range(to_load):
                pos = -(idx + 1)
                file = self._deque[pos][0]
                self._deque[pos] = (file, self.reader(file))

    def store(self, data):
        now = datetime.now(tz=timezone.utc)
        path = self._media_path(now)
        self.writer(data, path, **self.writer_opts)
        self.prune()
        if self.recent > 0:
            self._deque.append((path, data))
        else:
            self._deque.append((path, None))

        # Clear stale recent entry
        if len(self._deque) > self.recent and self.recent > 0:
            pos = -(self.recent + 1)
            file, buffer = self._deque[pos]
            del buffer
            self._deque[pos] = (file, None)

        # Update latest copy
        latest = os.path.join(self.dir, f"latest.{self.suffix}")
        shutil.copy2(path, latest)

    def fetch_recent(self):
        result = list()
        to_fetch = min(len(self._deque), self.recent)
        for idx in range(to_fetch):
            pos = -(to_fetch + idx)
            result.append(self._deque[pos])
        return result

    def fetch_index(self, index):
        return self._deque[index]

    def prune(self):
        while len(self._deque) >= self.max_files:
            file, buffer = self._deque.popleft()
            del buffer
            self._remove(file)

    def _remove(self, path):
        if not os.path.isfile:
            raise RuntimeError(f"{path} does not exist, cannot remove")
        os.remove(path)

    def _media_path(self, dt):
        dstr = self._dt_to_iso(dt)
        return os.path.join(self.dir, f"{self.root}_{dstr}.{self.suffix}")

    def _media_time(self, path):
        file = os.path.basename(path)
        pat = re.compile(f"{self.root}_(.*)\\.{self.suffix}")
        mat = pat.match(file)
        iso = mat.group(1)
        return self._iso_to_dt(iso)

    def _dt_to_iso(self, dt):
        return dt.isoformat(sep="T", timespec="seconds")

    def _iso_to_dt(self, iso):
        return datetime.datetime.fromisoformat(iso)


class MotionDetector:
    """Class to process images in sequence and look for changes.

    This uses the (stateful) helper tools from OpenCV to detect when an
    image contains changes from previous ones.

    Args:
        blur (int):  The odd number of pixels for blurring width
        threshold (int):  The grayscale threshold (0-255) for considering
            image changes in the blurred images.
        dilation (int):  The dilation iterations on the thresholded image.
        min_frac (float):  The minimum fraction of pixels that must change
            to count as a detection.
        max_frac (float):  The maximum fraction of pixels that can change
            to still count as a detection (rather than a processing error).

    """

    def __init__(
        self,
        blur=51,
        threshold=100,
        dilation=2,
        min_frac=0.005,
        max_frac=0.9,
    ):
        if blur % 2 == 0:
            raise ValueError("blur must be an odd integer")
        self.blur = blur
        if threshold < 0 or threshold > 255:
            raise ValueError("threshold should be between 0-255")
        self.thresh = int(threshold)
        if dilation > 10:
            raise ValueError("dilation should be a small, positive integer")
        self.dilation = dilation
        self.min_frac = min_frac
        self.max_frac = max_frac
        self._backsub = cv2.createBackgroundSubtractorMOG2()

    def reset(self):
        """Reset the background subtraction."""
        self._backsub = cv2.createBackgroundSubtractorMOG2()

    def process(self, data, skip=False):
        """Process image data and look for changes.

        Args:
            data (array):  The image data.
            skip (bool):  If True, accumulate image to background model, but
                do not look for motion.

        Returns:
            (tuple):  The (image, detection).

        """
        gray = cv2.cvtColor(data, cv2.COLOR_BGR2GRAY)
        blurred = cv2.GaussianBlur(gray, (self.blur, self.blur), 0)

        mask = self._backsub.apply(blurred)
        if skip:
            return (data, False)

        thresholded = cv2.threshold(mask, self.thresh, 255, cv2.THRESH_BINARY)[1]
        dilated = cv2.dilate(thresholded, None, iterations=self.dilation)
        cnts = cv2.findContours(
            dilated.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )
        cnts = imutils.grab_contours(cnts)

        # loop over the contours
        img_area = thresholded.shape[0] * thresholded.shape[1]
        min_area = self.min_frac * img_area
        max_area = self.max_frac * img_area
        detection = False
        for c in cnts:
            # if the contour is too small or too large, ignore it
            area = cv2.contourArea(c)
            if area < min_area:
                continue
            if area > max_area:
                continue
            # compute the bounding box for the contour, draw it on the frame,
            # and update the text
            detection = True
            (x, y, w, h) = cv2.boundingRect(c)
            cv2.rectangle(data, (x, y), (x + w, y + h), (0, 255, 0), 2)
        return (data, detection)


class FakeCamera:
    """Class used generate image data on demand for testing.

    The API for this class is intended to match the one provided
    by cv2.VideoCapture objects.

    Args:
        width (int):  Width of each frame in pixels.
        height (int):  Height of each frame in pixels.
        fps (float):  The frames per second of the stream.

    """

    def __init__(self, width=1280, height=720, fps=20.0):
        self.width = width
        self.height = height
        self.fps = fps

        # The number of frames to hold fixed before switching
        self.fixed_frames = int(60 * self.fps)

        # The size of a square to randomly place in the field of view
        self.square_dim = 100
        self.sq_half = self.square_dim // 2
        self._random_sq_pos()

        self.current = None
        self.frame_count = 0

    def isOpened(self):
        return True

    def _random_sq_pos(self):
        sq_y = int(self.height * np.random.random_sample(size=1)[0])
        sq_x = int(self.width * np.random.random_sample(size=1)[0])
        if sq_y > self.height - self.sq_half:
            sq_y = self.height - self.sq_half
        if sq_y < self.sq_half:
            sq_y = self.sq_half
        if sq_x > self.width - self.sq_half:
            sq_x = self.width - self.sq_half
        if sq_x < self.sq_half:
            sq_x = self.sq_half
        self.sq_x = sq_x
        self.sq_y = sq_y

    def grab(self):
        img = np.zeros((self.height, self.width, 3), dtype=np.uint8)
        img[:, :, 0] = 127
        img[:, :, 1] = 127
        img[:, :, 2] = 127
        if self.frame_count % self.fixed_frames == 0:
            self._random_sq_pos()
        img[
            self.sq_y - self.sq_half: self.sq_y + self.sq_half,
            self.sq_x - self.sq_half: self.sq_x + self.sq_half,
            0,
        ] = 255
        self.current = img
        self.frame_count += 1
        return True

    def retrieve(self):
        img = np.array(self.current)
        self.current = None
        return True, img

    def read(self):
        _ = self.grab()
        return self.retrieve()

    def release(self):
        pass



# Built-in python modules
import time as tm

# Control modules
from socs.common import moxa_serial as mx


class DWL:
    """
    The DWL object is for communicating with the DWL-5000XY gravity sensor

    Args:
    tcp_ip (str): TCP IP address
    tcp_port (int): TCP port
    """
    waittime = 0.05  # sec

    def __init__(self, tcp_ip=None, tcp_port=None, timeout=None, isSingle=False, verbose=0):
        self.tcp_ip = tcp_ip
        self.tcp_port = tcp_port
        self.isSingle = isSingle
        self.verbose = verbose

        # Connect to device
        msg = self.__conn(tcp_ip, tcp_port, timeout)
        print(msg)

    def __del__(self):
        print(f"Disconnecting from TCP IP {self.tcp_ip} at port {self.tcp_port}")
        self.ser.sock.close()
        return

    def get_angle(self):
        """ Measure the single-axis or two-axis angle """
        self.clean_serial()
        if self.isSingle:
            command = b"\x06\x01\x01\xAA\x00\x00\x00\x00\x00\x00\x00\x00"
        else:
            command = b"\x06\x01\x02\xAA\x00\x00\x00\x00\x00\x00\x00\x00"
        if self.verbose > 0:
            print(f'get_angle() command = {command}')

        read = []
        SIZE = 12

        # write and read serial
        self.ser.write(command)
        read_hex = self.ser.read(SIZE)
        read = [hex(r) for r in read_hex]
        if self.verbose > 0:
            print(f'read_hex = {read_hex}')
            print(f'read = {read}')

        # check the size of the string read
        if not len(read) == SIZE:
            msg = 'The size of the string read does not match with the expected size 12.'
            raise ValueError(msg)

        # check header matching and calculate the angles
        if self.isSingle:
            header = ['0x61', '0x11']
        else:
            header = ['0x61', '0x22']
        if read[0:2] == header:
            readInt = []
            val = ()
            for c in read:
                readInt.append((int)(c, 16))
            if self.isSingle:
                nums = [readInt[5], readInt[4], readInt[3], readInt[2]]
                angleX = (nums[0] << 24) + (nums[1] << 16) + (nums[2] << 8) + (nums[3])
                angleX = (angleX - 1800000) / 10000.
                val = (angleX)
                msg = f"Measured angle (1-axis) = {val}"
                if self.verbose > 0:
                    print(readInt)
                    print(nums)
                    print((nums[1] << 16) / 1e+4, (nums[2] << 8) / 1e+4, (nums[3]) / 1e+4)
                    print('angle X = {}'.format(angleX))
            else:
                readInt1 = readInt[5:8]
                readInt2 = readInt[2:5]
                readInt11 = readInt1
                readInt12 = readInt2
                numsX = [readInt11[2], readInt11[1], readInt11[0]]
                numsY = [readInt12[2], readInt12[1], readInt12[0]]
                angleX = (numsX[0] << 16) + (numsX[1] << 8) + (numsX[2])
                angleX = (angleX - 300000) / 10000.
                angleY = (numsY[0] << 16) + (numsY[1] << 8) + (numsY[2])
                angleY = (angleY - 300000) / 10000.
                val = (angleX, angleY)
                msg = f"Measured angle (2-axis) = {val}"
                if self.verbose > 0:
                    print(readInt)
                    print('numsX', numsX)
                    print((numsX[0] << 16) / 1e+4, (numsX[1] << 8) / 1e+4, (numsX[2]) / 1e+4)
                    print('numsY', numsY)
                    print((numsY[0] << 16) / 1e+4, (numsY[1] << 8) / 1e+4, (numsY[2]) / 1e+4)
                    print('angle X = {angleX}')
                    print('angle Y = {angleY}')
        else:
            msg = 'The header of the tilt sensor data is NOT matching.'
            raise RuntimeError(msg)
        return msg, val

    # ***** Helper Methods *****
    def __conn(self, tcp_ip=None, tcp_port=None, timeout=None):
        """
        Connect to the tilt sensor module

        Args:
        tcp_ip (str): TCP IP address
        tcp_port (int): TCP port
        """
        if tcp_ip is None or tcp_port is None:
            raise Exception(
                "Aborted DWL._conn() due to no "
                "TCP port specified")
        elif tcp_ip is not None and tcp_port is not None:
            self.ser = mx.Serial_TCPServer((tcp_ip, tcp_port), timeout, encoded=False)
            self.tcp_ip = tcp_ip
            self.tcp_port = int(tcp_port)
            self.using_tcp = True
            msg = "Connected to TCP IP %s at port %d" % (tcp_ip, tcp_port)
            command = b"\x06\x24\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"  # initialization command
            if self.verbose > 0:
                print('initialization command = {}'.format(command))
            self.ser.write(command)
            self.wait()
        else:
            raise Exception(
                "Aborted DWL._conn() due to unknown error")
        return msg

    def wait(self):
        """ Sleep """
        tm.sleep(self.waittime)
        return True

    def clean_serial(self):
        """ Flush the serial buffer """
        self.ser.flushInput()
        self.wait()
        return True



# HVAC fields for the LAT
#
# While most fields are listed explicitly in soaculib, the HVAC fields
# are parsed from the dataset on first read, because we also want to
# use regex to classify them for ocs-web / commanding.

import re
from dataclasses import dataclass

# Feed groups:

# hvac_data: readings (from thermometers), intended to only be
# recorded at low rate.
_DATA = 'hvac_data'

# hvac_ctrl: set points and on/off setting indicators.
_CTRL = 'hvac_ctrl'

# hvac_faults: faults and other flags.
_FAULTS = 'hvac_faults'


HVAC_SCHEMA = [
    {'type': 'ignore',
     'regex': r'Time|Year',
     'name_pattern': '{sname}',
     },

    {'type': 'T_indexed',
     'regex': r'Temperature (?P<loc>.*) (?P<idx>\d+)',
     'name_pattern': 'Temp_{sloc}_{idx}',
     'feed_group': _DATA,
     },

    {'type': 'T_average',
     'regex': r'Temperature Average (?P<loc>.*)',
     'name_pattern': 'Temp_{sloc}_avg',
     'feed_group': _DATA,
     },

    {'type': 'T_setpoint',
     'regex': r'Setpoint Temperature (?P<loc>.*)',
     'name_pattern': 'Temp_{sloc}_setpoint',
     'feed_group': _CTRL,
     },

    {'type': 'fan_on',
     'regex': r'Fan (?P<loc>.*) on',
     'name_pattern': 'Fan_{sloc}_on',
     'feed_group': _CTRL,
     },

    {'type': 'fan_fault',
     'regex': r'Fan (?P<loc>.*) Failure',
     'name_pattern': 'Fan_{sloc}_fault',
     'feed_group': _FAULTS,
     },

    {'type': 'fan_setpoint',
     'regex': r'Setpoint Speed Fan (?P<loc>.*)',
     'name_pattern': 'Fan_{sloc}_setpoint',
     'feed_group': _CTRL,
     },

    {'type': 'booster_on',
     'regex': r'Booster (?P<loc>.*) on',
     'name_pattern': 'Booster_{sloc}_on',
     'feed_group': _CTRL,
     },
    {'type': 'booster_fault',
     'regex': r'Booster (?P<loc>.*) Failure',
     'name_pattern': 'Booster_{sloc}_fault',
     'feed_group': _FAULTS,
     },

    {'type': 'heater_on',
     'regex': r'Heater on',
     'name_pattern': 'Heater_on',
     'feed_group': _CTRL,
     },

    {'type': 'unclassified',
     'regex': r'.+',
     'name_pattern': '{sname}',
     },
]


@dataclass
class HvacItem:
    """Describes the type, name, shortened name, and feed details for
    a single entry in DataSets.HVAC.

    """
    type: str
    name: str
    acu_name: str
    feed_group: str
    data: str


class HvacManager:
    """Interface class for collecting info about HVAC status fields
    from ACU.

    For now this assists with analyzing the HVAC dataset to classify
    the fields and assign them to the appropriate feed group.  Those
    are achieved by instantiating an instance (with no args) and then
    calling ``parse_fields`` followed by ``get_block_info``.

    In the future it may also help with generating control commands.

    """

    #: dict from field type to list of HvacItem.
    grouped_fields = None

    #: dict from ACU field name to the HvacItem.
    field_map = None

    def parse_fields(self, data):
        """Given the output from DataSets.HVAC, analyzes the fields
        therein and populates self.grouped_fields and self.field_map.
        This should be called on first valid data retrieval -- it will
        fail if self.grouped_fields has been populated already.

        """
        assert self.grouped_fields is None

        fields = {r['type']: [] for r in HVAC_SCHEMA}
        field_map = {}
        for f in data.keys():
            for sch in HVAC_SCHEMA:
                if m := re.match(sch['regex'], f):
                    data = m.groupdict()
                    if 'loc' in data:
                        data['sloc'] = data['loc'].replace(' ', '')
                    data['sname'] = f.replace(' ', '')
                    name = sch['name_pattern'].format(**data)
                    hv = HvacItem(type=sch['type'],
                                  name=name,
                                  acu_name=f,
                                  feed_group=sch.get('feed_group'),
                                  data=data)
                    fields[sch['type']].append(hv)
                    field_map[f] = hv
                    break
        self.grouped_fields = fields
        self.field_map = field_map

    def get_block_info(self):
        """Provides block info for monitor() process.  To be called
        after parse_fields.  Returns a map from acu_key to
        (feed_group, block_name, block_key).

        """
        output = {}
        for fields in self.grouped_fields.values():
            for f in fields:
                if f.feed_group:
                    output[f.acu_name] = (f.feed_group, 'ACU_' + f.feed_group, f.name)
        return output



import time

import serial

patterns = {
    'off': [0, 0, 0, 0, 0, 0],
    'on_1': [1, 1, 1, 0, 0, 0],
    'on_2': [1, 1, 1, 1, 1, 1],
    'stop': [0, 1, 1, 1, 0, 0],
}


class PCU:
    """Class to communicate with the phase compensation unit.

    Args:
        port (str): Path to USB device in '/dev/'

    Attributes:
        status (str): The status of the unit (off/on_1/on_2/stop)
    """

    def __init__(self, port):
        self.port = serial.Serial(
            port,
            baudrate=19200,
            timeout=1,
            bytesize=serial.EIGHTBITS,
            stopbits=serial.STOPBITS_ONE,
        )

    def close(self):
        self.port.close()

    def sleep(self):
        time.sleep(0.1)

    def read(self):
        return self.port.read_until(b'\n\r').strip().decode()

    def clear_buffer(self):
        while True:
            res = self.read()
            if len(res) == 0:
                break

    def relay_on(self, channel):
        cmd = "relay on " + str(channel) + "\n\r"
        self.port.write(cmd.encode('utf-8'))
        self.sleep()

    def relay_off(self, channel):
        cmd = "relay off " + str(channel) + "\n\r"
        self.port.write(cmd.encode('utf-8'))
        self.sleep()

    def relay_read(self, channel):
        cmd = "relay read " + str(channel) + "\n\r"
        self.port.write(cmd.encode('utf-8'))
        self.sleep()
        response = self.read()
        response = self.read()
        if response == "on":
            return 1
        elif response == "off":
            return 0
        else:
            return -1

    def send_command(self, command):
        pattern = patterns[command]
        for i, p in zip([0, 1, 2, 5, 6, 7], pattern):
            if p:
                self.relay_on(i)
            else:
                self.relay_off(i)
            self.sleep()
            self.read()

    def get_status(self):
        """get_status()

        **Task** - Get the operation mode of the phase compensation unit.
        off: The compensation phase is zero.
        on_1: The compensation phase is +120 deg.
        on_2: The compensation phase is -120 deg.
        stop: Stop the HWP spin.
        """
        channel = [0, 1, 2, 5, 6, 7]
        channel_switch = []

        for i in channel:
            channel_switch.append(self.relay_read(i))
        if -1 in channel_switch:
            return 'failed'
        for command, pattern in patterns.items():
            if channel_switch == pattern:
                return command
        return 'undefined'



#!/usr/bin/env python3
"""OCS agent for stimulator encoder
"""
import argparse
import os
import time

import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock

from socs.agents.stimulator_encoder.drivers import (PATH_LOCK, StimEncReader,
                                                    get_path_dev)


class StimEncAgent:
    """OCS agent class for stimulator encoder

    Parameters
    ----------
    path_dev : str or pathlib.Path
        Path to the generic-uio device file for str_rd IP.
    path_lock : str or pathlib.Path
        Path to the lockfile.
    """

    def __init__(self, agent, path_dev=None, path_lock=PATH_LOCK):
        self.active = True
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()
        self.take_data = False

        if path_dev is None:
            self._path_dev = get_path_dev()

        self._dev = StimEncReader(self._path_dev, path_lock, verbose=False)

        self.initialized = False

        agg_params = {'frame_length': 60, 'exclude_influx': True}
        self.agent.register_feed('stim_enc',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

        agg_params_downsampled = {'frame_length': 60}
        self.agent.register_feed('stim_enc_downsampled',
                                 record=True,
                                 agg_params=agg_params_downsampled,
                                 buffer_time=1)

    def acq(self, session, params):
        """acq()

        **Process** - Start acquiring data.

        Notes
        -----
        An example of the session data::

            >>> response.session['data']

            {'timestamp_tai': 1736541833.679634,
             'state': 1,
             'freq': 10.1,
             'timestamp': 1736541796.779634
            }
        """
        pace_maker = Pacemaker(0.5)

        with self.lock.acquire_timeout(timeout=0, job='acq') as acquired:
            if not acquired:
                self.log.warn(
                    f'Could not start acq because {self.lock.job} is already running')
                return False, 'Could not acquire lock.'

            self.take_data = True
            session.data = {}

            self._dev.run()

            tai_latest = 0.
            en_st_latest = 0
            pulse_count = 0

            time_prev = time.time()
            while self.take_data:
                # Data acquisition
                time_now = time.time()
                data = {'block_name': 'stim_enc', 'data': {}}

                tai_list = []
                ts_list = []
                en_st_list = []

                while not self._dev.fifo.empty():
                    _d = self._dev.fifo.get()
                    tai_list.append(_d.time.tai)
                    ts_list.append(_d.utime)
                    en_st_list.append(_d.state)

                if len(tai_list) != 0:
                    data['timestamps'] = ts_list
                    data['data']['timestamps_tai'] = tai_list
                    data['data']['state'] = en_st_list
                    self.agent.publish_to_feed('stim_enc', data)

                    tai_latest = tai_list[-1]
                    en_st_latest = en_st_list[-1]
                    pulse_count += len(en_st_list)

                # Slow feed for InfluxDB and session.data
                freq = pulse_count / (time_now - time_prev)
                field_dict = {'timestamp_tai': tai_latest,
                              'state': en_st_latest,
                              'freq': freq,
                              'timestamp': time_now}
                session.data.update(field_dict)

                data_downsampled = {'timestamp': time_now,
                                    'block_name': 'stim_enc_downsampled',
                                    'data': {
                                        'timestamps_tai': tai_latest,
                                        'state': en_st_latest,
                                        'freq': freq
                                    }}
                self.agent.publish_to_feed('stim_enc_downsampled',
                                           data_downsampled)
                time_prev = time_now
                pulse_count = 0

                pace_maker.sleep()

        self.agent.feeds['stim_enc'].flush_buffer()
        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        Stops the data acquisiton.
        """
        if self.take_data:
            self.take_data = False
            self._dev.stop()
            return True, 'requested to stop taking data.'

        return False, 'acq is not currently running.'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--path-dev', default=None, type=str,
                        help='Path to the device file.')
    pgroup.add_argument('--path-lock', default=PATH_LOCK, type=str,
                        help='Path to the lock file.')

    return parser


def main(args=None):
    """Boot OCS agent"""
    txaio.start_logging(level=os.environ.get('LOGLEVEL', 'info'))

    parser = make_parser()
    args = site_config.parse_args('StimEncAgent',
                                  parser=parser,
                                  args=args)

    agent_inst, runner = ocs_agent.init_site_agent(args)

    stim_enc_agent = StimEncAgent(agent_inst)

    agent_inst.register_process(
        'acq',
        stim_enc_agent.acq,
        stim_enc_agent._stop_acq,
        startup=True
    )

    runner.run(agent_inst, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import os
import socket
import time

import numpy
import txaio
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock


class LDMonitor:
    """Receives and decodes data of the lightning detector via UDP

    Parameters
    ----------
    host : str
        Address of the computer reading the data.
    port : int
        Port of host where data will be received, default 1110.

    Attributes
    ----------
    port : int
        Port number in the local host to be bound to receive the data
    log : txaio.tx.Logger
        txaio logger object, created by the OCSAgent
    sockopen : bool
        Indicates when the socket is open
    inittime : float
        Logs the time at which initialization was carried out
    data_dict : dictionary
        Dictionary data stored from the lightning detector
    newdata_dict : dictionary
        Dictionary where new data is received
    """

    def __init__(self, port=1110):
        self.port = port
        self.log = txaio.make_logger()

        # check if socket has been opened
        if hasattr(self, 'sockopen'):
            if self.sockopen:
                self.sock.close()
                self.log.info('Socket closed preemptively')

        # open and bing socket to receieve lightning detector data
        try:
            self.log.info('Opening socket')
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        except BaseException:
            self.log.info('Failed to create socket')

        try:
            self.log.info('Binding socket')
            self.sock.bind(('', self.port))
            self.sockopen = True
            self.inittime = time.time()
        except BaseException:
            self.log.info('Failed to bind socket')

        # initialize variables to account for absence of previous data
        self.data_dict = {
            'd_type': numpy.nan,
            'field_value': numpy.nan,
            'rot_fault': 0,
            'time_last': -1.,
            'tsince_last': -1.,
            'dist': -1,
            'unit_d': 0,
            'high_field': -1,
            'hifield_value': -1000.,
            'alarm_r': 0,
            'alarm_o': 0,
            'alarm_y': 0,
            'delay_g': 1,
            'clear': 0,
            'r_timer': 0,
            'o_timer': 0,
            'y_timer': 0,
            'g_timer': 0,
            'allclear_timer': 0,
            'faultcode': 0
        }

        self.log.info('LDMonitor function initialized')

    def read_data(self):
        """
        Receives data from the lightning detector via UDP,
        and returns a dictionary containing formatted data
        """

        self.data, _ = self.sock.recvfrom(1024)
        self.data = self.data.decode('utf-8')

        # "e-field" sentence
        if self.data[0] == '$':
            data_split = self.data[1:].split(',')
            rot_fault = int(data_split[1].split('*')[0])
            self.newdata_dict = {
                'd_type': 0,
                'field_value': float(data_split[0]),
                'rot_fault': rot_fault
            }
            self.data_dict.update(self.newdata_dict)
            return self.data_dict

        elif self.data[0] == '@':
            param = self.data[1:3]

            # "lightning strike" sentence
            if param == 'LI':
                data_split = self.data.split(',')[1:]
                if data_split[1].split('*')[0] == 'Miles':
                    unit_d = 0
                elif data_split[1].split('*')[0] == 'Km':
                    unit_d = 1

                self.newdata_dict = {
                    'd_type': 1,
                    'time_last': time.time(),
                    'dist': int(data_split[0]),
                    'unit_d': unit_d
                }
                self.data_dict.update(self.newdata_dict)

                self.log.info('Lightning strike detected!')

                return self.data_dict

            # "high e-field" sentence, account for 2 types
            elif param == 'HF':
                data_split = self.data[1:].split(',')
                if len(data_split) == 1:
                    self.newdata_dict = {
                        'd_type': 2,
                        'high_field': 1,
                        'hifield_value': float(self.data_dict['field_value'])
                    }
                else:
                    self.newdata_dict = {
                        'd_type': 2,
                        'hifield_value': float(data_split[1])
                    }
                self.data_dict.update(self.newdata_dict)
                return self.data_dict

            # "status" sentence
            elif param == 'ST':
                faultcode = int(self.data.split(',')[-1].split('*')[0], 16)
                data_split = [int(i) for i in self.data.split(',')[1:-1]]

                self.newdata_dict = {
                    'd_type': 3,
                    'alarm_r': data_split[0],
                    'alarm_o': data_split[1],
                    'alarm_y': data_split[2],
                    'delay_g': data_split[3],
                    'clear': data_split[4],
                    'r_timer': data_split[5],
                    'o_timer': data_split[6],
                    'y_timer': data_split[7],
                    'g_timer': data_split[8],
                    'allclear_timer': data_split[9],
                    'faultcode': faultcode
                }

                self.data_dict.update(self.newdata_dict)
                return self.data_dict

            # disregard "alarm timers" sentence, but update sentence type
            elif param == 'WT':
                self.newdata_dict = {'d_type': 4}
                self.data_dict.update(self.newdata_dict)
                return self.data_dict

    def read_cycle(self):
        """
        In each cycle data is read and then parsed following
        the format required to publish data to the ocs feed
        """
        try:
            self.read_data()

            # updates time since last strike if strike data exists
            if self.data_dict['time_last'] == -1.:
                self.data_dict['tsince_last'] = -1.
            else:
                self.data_dict['tsince_last'] = (time.time()
                                                 - self.data_dict['time_last'])

            return self.data_dict

        except BaseException:
            pass
            self.log.info('LD data read error, passing to next data iteration')


class LDMonitorAgent:
    """Monitor the Lightning Detector data via UDP.

    Parameters
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    sample_interval : float
        Time between samples in seconds.

    Attributes
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    take_data : bool
        Tracks whether or not the agent is actively issuing SNMP GET commands
        to the ibootbar. Setting to false stops sending commands.
    log : txaio.tx.Logger
        txaio logger object, created by the OCSAgent
    """

    def __init__(self, agent, sample_interval=15.):

        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.pacemaker_freq = 1. / sample_interval

        self.initialized = False
        self.take_data = False

        self.LDMonitor = None

        agg_params = {
            'frame_length': 10 * 60  # [sec]
        }
        self.agent.register_feed('ld_monitor',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

    def _connect(self):
        """connect()
        Instantiates LD object and check if client is open
        """

        self.LDMonitor = LDMonitor()
        self.initialized = True

    @ocs_agent.param('auto_acquire', default=False, type=bool)
    def init_ld_monitor(self, session, params=None):
        """init_ld_monitor(auto_acquire=False)

        **Task** - Perform first time setup of the LD.

        Parameters:
            auto_acquire (bool, optional): Starts data acquisition after
                initialization if True. Defaults to False.

        """

        if self.initialized:
            return True, "Already initialized."

        with self.lock.acquire_timeout(3, job='init') as acquired:
            if not acquired:
                self.log.warn("Could not start init because "
                              "{} is already running".format(self.lock.job))
                return False, "Could not acquire lock."

            self._connect()
            if not self.initialized:
                return False, 'Could not connect to LD'

        # Start data acquisition if requested
        if params['auto_acquire']:
            self.agent.start('acq')
        return True, 'LD initialized.'

    @ocs_agent.param('_')
    def acq(self, session, params=None):
        """acq()

        **Process** - Starts the data acquisition process

        Notes
        _____
        The most recent data collected is stored in session data in the
        structure::

            >>> response.session['data']
            {'fields':{
                'd_type': 3, 'field_value': 0.28, 'rot_fault': 0,
                'time_last': -1.0, 'tsince_last': -1.0, 'dist': -1, 'unit_d': 0,
                'high_field': -1, 'hifield_value': -1000.0, 'alarm_r': 0,
                'alarm_o': 0, 'alarm_y': 0, 'delay_g': 1, 'clear': 1, 'r_timer': 0,
                'o_timer': 0, 'y_timer': 0, 'g_timer': 0, 'allclear_timer': 0,
                'faultcode': 0
                },
                ...
                'connection': {
                    'conn_timestamp': 1711285858.1063662,
                    'connected': True}, 'data_timestamp': 1711285864.6254003
                    }
            }

        """
        with self.lock.acquire_timeout(0, job='acq') as acquired:
            if not acquired:
                self.log.warn("Could not start acq because {} is already running"
                              .format(self.lock.job))
                return False, "Could not acquire lock."

            self.take_data = True

            session.data = {"fields": {}}

            pm = Pacemaker(self.pacemaker_freq)
            while self.take_data:
                pm.sleep()

                current_time = time.time()
                data = {
                    'timestamp': current_time,
                    'connection': {},
                    'data': {}
                }
                if not self.LDMonitor.sockopen:
                    self.initialized = False

                # Try to re-initialize if connection lost
                if not self.initialized:
                    self._connect()

                # Only get readings if connected
                if self.initialized:
                    session.data.update({'connection': {'conn_timestamp': self.LDMonitor.inittime,
                                                        'connected': True}})

                    ld_data = self.LDMonitor.read_cycle()

                    if ld_data:
                        for key, value in ld_data.items():
                            data['data'][key] = value
                        session.data.update({'data_timestamp': current_time,
                                             'fields': ld_data})
                        self.log.debug(ld_data)
                    else:
                        self.log.info('Connection error or error in processing data.')
                        self.initialized = False

                # Continue trying to connect
                if not self.initialized:
                    session.data.update({'connection': {'last_attempt': time.time(),
                                                        'connected': False}})
                    self.log.info('Trying to reconnect.')
                    continue

                for field, val in data['data'].items():
                    _data = {
                        'timestamp': current_time,
                        'block_name': field,
                        'data': {field: val}
                    }
                    self.agent.publish_to_feed('ld_monitor', _data)

            self.agent.feeds['ld_monitor'].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params=None):
        """
        Stops acq process.
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'


def make_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--mode', type=str, choices=['idle', 'init', 'acq'],
                        help="Starting action for the agent.")
    pgroup.add_argument("--sample-interval", type=float, default=.2, help="Time between samples in seconds.")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = make_parser()

    # Interpret options in the context of site_config.
    args = site_config.parse_args(agent_class='LDMonitor',
                                  parser=parser,
                                  args=args)

    # Automatically acquire data if requested (default)
    init_params = False
    if args.mode == 'init':
        init_params = {'auto_acquire': False}
    elif args.mode == 'acq':
        init_params = {'auto_acquire': True}
    print('init_params', init_params)
    agent, runner = ocs_agent.init_site_agent(args)

    p = LDMonitorAgent(agent, sample_interval=args.sample_interval)
    agent.register_task('init_ld_monitor', p.init_ld_monitor,
                        startup=init_params)
    agent.register_process('acq', p.acq, p._stop_acq)
    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



"""
HWP Emulation module
"""

import argparse
import logging
import pickle as pkl
import socket
import threading
import time
import traceback as tb
from dataclasses import asdict, dataclass, field
from typing import Dict, List, Optional, Tuple

import numpy as np
import pytest

from socs.agents.hwp_pid.drivers.pid_controller import PID
from socs.testing import device_emulator


def _find_open_port() -> int:
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def hex_str_to_dec(hex_value, decimal=3):
    """Converts a hex string to a decimal float"""
    return float(int(hex_value, 16)) / 10**decimal


@dataclass
class PMXState:
    """
    State of the PMX Emulator
    """

    output: bool = False
    current: float = 0
    current_limit: float = 10.0
    voltage_limit: float = 10.0
    voltage: float = 0
    source: str = "volt"


@dataclass
class PIDState:
    """State of the PID Emulator"""

    direction: str = "forward"
    freq_setpoint: float = 0.0


NUM_PCU_RELAYS = 8


@dataclass
class PCUState:
    """State of PCU emulator"""

    relays: List[bool] = field(
        default_factory=lambda: [False for _ in range(NUM_PCU_RELAYS)]
    )


# Gripper state information, taken from sobonelib: https://github.com/simonsobs/sobonelib/blob/main/hwp_gripper/control/state_monitor.py


COLD_LIMIT_POS = 13.0  # mm
WARM_LIMIT_POS = 10.0  # mm
LIMIT_TOLERANCE = 1.0  # mm


@dataclass
class LimitState:
    state: bool = False


@dataclass
class ActuatorState:
    axis: int
    limits: Dict[str, LimitState] = field(
        default_factory=lambda: {
            "cold_grip": LimitState(),
            "warm_grip": LimitState(),
        }
    )
    pos: float = 0
    brake: bool = False
    emg: bool = False

    def update(self) -> None:
        self.limits["cold_grip"].state = self.pos >= COLD_LIMIT_POS - LIMIT_TOLERANCE
        self.limits["warm_grip"].state = self.pos >= WARM_LIMIT_POS - LIMIT_TOLERANCE


@dataclass
class JXCState:
    setup: bool = False
    svon: bool = False
    busy: bool = False
    seton: bool = False
    inp: bool = False
    svre: bool = False
    alarm: bool = False
    out: int = 0

    @property
    def status(self):
        """
        Current status of the controller determined from JXC input pins.
        Described on page 43 of the JXC manual:
            https://www.smcworld.com/assets/manual/en-jp/files/SFOD-OMT0010.pdf
        """

        # Bit string that we can use to easily check current state
        status_bits = [self.busy, self.inp, self.svre, self.seton]
        bit_rep = "".join([str(int(s)) for s in status_bits])

        if self.out == 0:
            if bit_rep == "0000":
                return "powered_down_servo_off"
            elif bit_rep == "0010":
                return "powered_down_servo_on"
            elif bit_rep == "0100":
                return "return_to_origin"
            elif bit_rep == "0111":
                return "home"
        else:
            # This contains a number of different cases, but from the docs they
            # don't seem to be 1-to-1 to bit_rep
            return f"output_step_{self.out}"


@dataclass
class GripperState:
    actuators: Tuple[ActuatorState, ActuatorState, ActuatorState] = field(
        default_factory=lambda: (
            ActuatorState(axis=1),
            ActuatorState(axis=2),
            ActuatorState(axis=3),
        )
    )
    jxc: JXCState = field(default_factory=JXCState)
    last_packet_received: float = 0.0
    last_limit_received: float = 0.0
    last_encoder_received: float = 0.0
    calibrated: bool = False
    calibrated_at: float = 0
    powered: bool = True
    # If is_cold, use cold limit switch to stop pos, else use warm limit
    is_cold: bool = False

    _expiration_time: float = 10.0
    _last_enc_state: Optional[int] = None
    _left_home: bool = False

    @property
    def expired(self):
        """Returns True if the gripper state has not been updated in a while"""
        return time.time() - self.last_packet_received > self._expiration_time

    def update(self) -> None:
        now = time.time()
        self.last_packet_received = now
        self.last_limit_received = now
        self.last_encoder_received = now
        for act in self.actuators:
            act.update()


@dataclass
class HWPState:
    """State of the HWP Emulator"""

    cur_freq: float = 0.0
    pmx: PMXState = field(default_factory=PMXState)
    pid: PIDState = field(default_factory=PIDState)
    pcu: PCUState = field(default_factory=PCUState)
    gripper: GripperState = field(default_factory=GripperState)
    lock = threading.Lock()


def _create_logger(name, log_level=logging.INFO):
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    if len(logger.handlers) == 0:
        formatter = logging.Formatter("%(name)s: %(message)s")
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger


def lerp(start, end, t):
    return (1 - t) * start + t * end


BB_CLOCK_FREQ = 200e6  # Hz
NUM_SLITS = 570  # Slits per rotation
ENC_COUNTER_LEN = 120


class HWPEmulator:
    def __init__(
        self,
        pid_port=0,
        pmx_port=0,
        gripper_port=0,
        pcu_port="./responder",
        log_level=logging.INFO,
        enc_port=0,
        lerp_frac=0.9,
    ):
        self.pid_port = pid_port
        self.pmx_port = pmx_port
        self.pcu_port = pcu_port
        self.enc_port = enc_port
        self.gripper_port = gripper_port

        self.state = HWPState()
        self.lerp_frac = lerp_frac

        self.pid_device = device_emulator.DeviceEmulator([])
        self.pid_device.get_response = self.process_pid_msg
        self.pid_device.logger = _create_logger("PID", log_level=log_level)

        self.pmx_device = device_emulator.DeviceEmulator([])
        self.pmx_device.get_response = self.process_pmx_msg
        self.pmx_device.logger = _create_logger("PMX", log_level=log_level)

        self.pcu_device = device_emulator.DeviceEmulator([])
        self.pcu_device.get_response = self.process_pcu_msg
        self.pcu_device.logger = _create_logger("PCU", log_level=log_level)

        self.gripper_device = device_emulator.DeviceEmulator([])
        self.gripper_device.get_response = self.process_gripper_msg
        self.gripper_device.encoding = None
        self.gripper_device.logger = _create_logger("GRIP", log_level=log_level)

        self.update_thread = threading.Thread(target=self.update_loop)
        self.run_update = False
        self.enc_thread = threading.Thread(target=self.encoder_thread_func)
        self.run_enc_thread = False

        if self.enc_port == 0:
            self.enc_port = _find_open_port()

        self.logger = _create_logger("HWP", log_level=log_level)

    def start(self):
        """Start up TCP Sockets and update loop"""
        if self.pid_port is not None:
            self.pid_device.create_tcp_relay(self.pid_port)
        if self.pmx_port is not None:
            self.pmx_device.create_tcp_relay(self.pmx_port)
        if self.gripper_port is not None:
            self.gripper_device.create_tcp_relay(self.gripper_port)
        if self.pcu_port is not None:
            self.pcu_device.create_serial_relay()

        self.logger.info(f"PID port: {self.pid_device.socket_port}")
        self.logger.info(f"PMX port: {self.pmx_device.socket_port}")
        # self.logger.info(f"PCU port: {self.pcu_device}")
        self.logger.info(f"Gripper port: {self.gripper_device.socket_port}")

        self.update_thread.start()
        self.enc_thread.start()

    def shutdown(self):
        """Shutdown TCP Sockets and update loop"""
        self.run_update = False
        self.run_enc_thread = False
        self.update_thread.join()
        self.enc_thread.join()
        self.pid_device.shutdown()
        self.pmx_device.shutdown()
        self.pcu_device.shutdown()
        self.gripper_device.shutdown()
        self.logger.info("Finished shutdown")

    def encoder_thread_func(self):
        # Function that sends UDP packets to be interpreted by the encoder agent.
        # This contains info about encoder edges.
        self.logger.info("Starting encoder thread...")
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        addr = ("", self.enc_port)

        # Header and quad info
        header = np.array([0x1EAF, 1], dtype=np.uint32)
        edge_count = np.arange(ENC_COUNTER_LEN, dtype=np.uint32)
        clock_full = np.empty(edge_count.shape, dtype=np.uint64)

        last_clock = 0
        self.run_enc_thread = True
        while self.run_enc_thread:
            f = self.state.cur_freq
            if f < 0.01:
                time.sleep(0.1)
                continue
            sleep_time = ENC_COUNTER_LEN / (f * NUM_SLITS)
            time.sleep(sleep_time)

            if f < 0.01:
                time.sleep(0.05)
                continue
            else:
                clock_step = BB_CLOCK_FREQ / (2 * self.state.cur_freq * NUM_SLITS)
                next_clock = last_clock + ENC_COUNTER_LEN * clock_step
                clock_full = np.linspace(
                    last_clock, next_clock, ENC_COUNTER_LEN, dtype=np.uint64
                )
                last_clock = next_clock
                edge_count += ENC_COUNTER_LEN
                clock = (clock_full & 0xFFFFFFFF).astype(np.uint32)
                clock_overflow = (clock_full >> 32).astype(np.uint32)
                byte_data = np.hstack(
                    [header, clock, clock_overflow, edge_count]
                ).tobytes()
                sock.sendto(byte_data, addr)

                time.sleep(ENC_COUNTER_LEN / (2 * self.state.cur_freq * NUM_SLITS))
        self.logger.info("Stopping encoder thread")

    def update_loop(self) -> None:
        """Update HWP state"""
        self.run_update = True
        s: HWPState = self.state
        self.logger.info("Starting update thread")

        while self.run_update:
            with s.lock:
                if s.pmx.source == "volt":
                    s.cur_freq = lerp(s.cur_freq, s.pid.freq_setpoint, self.lerp_frac)
                s.gripper.update()
            time.sleep(0.2)
        self.logger.info("Stopping update thread")

    def process_pcu_msg(self, data) -> str:
        self.logger.debug(data)
        try:
            _, cmd, chan_str = data.split(" ")
            chan = int(chan_str)
        except Exception:
            raise ValueError(f"Unable to parse message: {data}")

        assert chan < NUM_PCU_RELAYS
        assert cmd in ["read", "on", "off"]

        if cmd == "read":
            if self.state.pcu.relays[chan]:
                return f"{chan}\n\ron\n\r"
            else:
                return f"{chan}\n\roff\n\r"
        elif cmd == "on":
            self.state.pcu.relays[chan] = 1
        elif cmd == "off":
            self.state.pcu.relays[chan] = 0
        return "\n\r"

    def process_pmx_msg(self, data):
        """Process messages for PMX emulator"""
        cmd = data.split(" ")[0].strip()
        self.logger.debug(cmd)
        with self.state.lock:
            # Output commands
            if cmd == "output":
                val = int(data.split(" ")[1].strip())
                self.logger.info("Setting output to %d", val)
                self.state.pmx.output = bool(val)
            elif cmd == "output:protection:clear":
                self.logger.info("Commanded to clear alarms")
            elif cmd == "output?":
                return str(int(self.state.pmx.output))

            # Current (limit) commands
            elif cmd == "curr":
                val = float(data.split(" ")[1].strip())
                self.logger.info("Setting current to %.3f", val)
                self.state.pmx.current = val
            elif cmd == "curr:prot":
                val = float(data.split(" ")[1].strip())
                self.logger.info("Setting current limit to %.3f", val)
                self.state.pmx.current_limit = val
            elif cmd == "curr?":
                return f"{self.state.pmx.current}\n"
            elif cmd == "curr:prot?":
                return f"{self.state.pmx.current_limit}\n"
            elif cmd == "meas:curr?":
                return f"{self.state.pmx.current}\n"

            # Voltage (limit) commands
            elif cmd == "volt":
                val = float(data.split(" ")[1].strip())
                self.logger.info("Setting current to %.3f", val)
                self.state.pmx.voltage = val
            elif cmd == "volt:prot":
                val = float(data.split(" ")[1].strip())
                self.logger.info("Setting voltage limit to %.3f", val)
                self.state.pmx.voltage_limit = val
            elif cmd == "volt:prot?":
                return f"{self.state.pmx.voltage_limit}\n"
            elif cmd == "volt?":
                return f"{self.state.pmx.voltage}\n"
            elif cmd == "meas:volt?":
                return f"{self.state.pmx.voltage}\n"

            # Error codes
            elif cmd == ":system:error?":  # Error codes
                return '0,"No error"\n'
            elif cmd == "stat:ques?":  # Status Codes
                return "0"
            elif cmd == "volt:ext:sour?":
                return f"{self.state.pmx.source}\n"
            else:
                self.logger.info("Unknown cmd: %s", data)
                if "?" in cmd:
                    return "unknown"

    def process_pid_msg(self, data):
        """Process messages for PID emulator"""
        logger = self.pid_device.logger
        cmd = data.split(" ")[0].strip()
        with self.state.lock:
            # self.logger.debug(cmd)
            if cmd == "*W02400000":
                self.state.pid.direction = "forward"
                logger.info("Setting direction: forward")
                return "asdfl"
            elif cmd == "*W02401388":
                self.state.pid.direction = "reverse"
                logger.info("Setting direction: reverse")
                return "asdfl"
            elif cmd.startswith("*W014"):
                setpt = hex_str_to_dec(cmd[5:], 3)
                logger.info("SETPOINT %s Hz", setpt)
                self.state.pid.freq_setpoint = setpt
                return "sdflsf"
            elif cmd == "*X01":  # Get frequency
                return f"X01{self.state.cur_freq:0.3f}"
            elif cmd == "*R01":  # Get Target
                return f"R0140{PID._convert_to_hex(self.state.pid.freq_setpoint, 3)}"
            elif cmd == "*R02":  # Get Direction
                if self.state.pid.direction == "forward":
                    return "R02400000"
                else:
                    return "R02401388"
            else:
                self.logger.info("Unknown cmd: %s", cmd)
                return "unknown"

    def process_gripper_msg(self, msg_bytes: bytes) -> bytes:
        msg = msg_bytes.decode()
        try:
            state: GripperState = self.state.gripper
            cmd = msg.strip().split(" ")
            self.logger.debug(f"Gripper: {msg}")

            def raise_error():
                raise ValueError(f"Unable to process msg: {msg}")

            if cmd[0] == "ON":
                state.powered = True
            elif cmd[0] == "OFF":
                state.powered = False
            elif cmd[0] == "HOME":
                for act in state.actuators:
                    act.pos = 0
            elif cmd[0] == "MOVE":
                act_idx = int(cmd[2]) - 1
                new_pos = state.actuators[act_idx].pos + float(cmd[3])
                if state.is_cold and new_pos >= COLD_LIMIT_POS:
                    state.actuators[act_idx].pos = COLD_LIMIT_POS
                    return pkl.dumps({"result": False, "log": []})
                elif (not state.is_cold) and new_pos >= WARM_LIMIT_POS:
                    state.actuators[act_idx].pos = WARM_LIMIT_POS
                    return pkl.dumps({"result": False, "log": []})
                else:
                    state.actuators[act_idx].pos = new_pos
                    return pkl.dumps({"result": True, "log": []})
            elif cmd[0] == "BRAKE":
                enabled = cmd[1] == "ON"
                if len(cmd) == 3:
                    act_idxs = [int(cmd[2]) - 1]
                else:
                    act_idxs = [0, 1, 2]
                for i in act_idxs:
                    state.actuators[i].brake = enabled
            elif cmd[0] == "EMG":
                enabled = cmd[1] == "ON"
                if len(cmd) == 3:
                    act_idxs = [int(cmd[2]) - 1]
                else:
                    act_idxs = [0, 1, 2]
                for i in act_idxs:
                    state.actuators[i].emg = enabled
            elif cmd[0] == "GET_STATE":
                return pkl.dumps({"result": asdict(state), "log": []})

            return pkl.dumps({"result": True, "log": []})
        except Exception:
            print(tb.format_exc())
            return b""


def create_hwp_emulator_fixture(**kwargs):
    """
    Creates a fixture for the HWP Emulator to use in tests.
    """

    @pytest.fixture()
    def create_emulator():
        em = HWPEmulator(**kwargs)
        try:
            em.start()
            yield em
        finally:
            em.shutdown()

    return create_emulator


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--pid-port", type=int, default=0)
    parser.add_argument("--pmx-port", type=int, default=0)
    args = parser.parse_args()

    hwp_em = HWPEmulator(pid_port=args.pid_port, pmx_port=args.pmx_port)
    try:
        hwp_em.start()
        while True:
            time.sleep(1)
    finally:
        hwp_em.shutdown()



import argparse
import time
from dataclasses import dataclass
from queue import Queue

import serial
import txaio
from twisted.internet import defer, reactor, threads

txaio.use_twisted()

from ocs import ocs_agent, site_config

import socs.agents.hwp_pcu.drivers.hwp_pcu as pcu


class Actions:
    class BaseAction:
        def __post_init__(self):
            self.deferred = defer.Deferred()
            self.log = txaio.make_logger()

    @dataclass
    class SendCommand (BaseAction):
        command: str


def process_action(action, PCU: pcu.PCU):
    """Process an action with PCU hardware"""
    if isinstance(action, Actions.SendCommand):
        PCU.send_command(action.command)
        action.log.info(f"Command: {action.command}")


class HWPPCUAgent:
    """Agent to phase compensation improve the CHWP motor efficiency

    Args:
        agent (ocs.ocs_agent.OCSAgent): Instantiated OCSAgent class for this agent
        port (str): Path to USB device in '/dev/'
    """

    def __init__(self, agent, port):
        self.agent: ocs_agent.OCSAgent = agent
        self.log = agent.log
        self.port = port
        self.action_queue = Queue()

        agg_params = {'frame_length': 60}
        self.agent.register_feed(
            'hwppcu', record=True, agg_params=agg_params)

    @defer.inlineCallbacks
    @ocs_agent.param('command', default='off', type=str, choices=['off', 'on_1', 'on_2', 'stop'])
    def send_command(self, session, params):
        """send_command(command)

        **Task** - Send commands to the phase compensation unit.
        off: The compensation phase is zero.
        on_1: The compensation phase is +120 deg.
        on_2: The compensation phase is -120 deg.
        stop: Stop the HWP spin.

        Parameters:
            command (str): set the operation mode from 'off', 'on_1', 'on_2' or 'stop'.

        """
        action = Actions.SendCommand(**params)
        self.action_queue.put(action)
        session.data = yield action.deferred
        return True, f"Set relays for cmd={action.command}"

    def _process_actions(self, PCU: pcu.PCU):
        while not self.action_queue.empty():
            action = self.action_queue.get()
            try:
                self.log.info(f"Running action {action}")
                res = process_action(action, PCU)
                reactor.callFromThread(action.deferred.callback, res)
            except Exception as e:
                self.log.error(f"Error processing action: {action}")
                reactor.callFromThread(action.deferred.errback, e)

    def _get_and_publish_data(self, PCU: pcu.PCU, session):
        now = time.time()
        data = {'timestamp': now,
                'block_name': 'hwppcu',
                'data': {}}
        status = PCU.get_status()

        data['data']['status'] = status
        self.agent.publish_to_feed('hwppcu', data)
        session.data = {'status': status, 'last_updated': now}

        if status in ['failed', 'undefined']:
            PCU.clear_buffer()
            self.log.warn(f'Status is {status}, cleared buffer')

    def _clear_queue(self):
        while not self.action_queue.empty():
            action = self.action_queue.get()
            action.deferred.errback(Exception("Action cancelled"))

    def main(self, session, params):
        """
        **Process** - Main process for PCU agent.
        """
        PCU = None

        threads.blockingCallFromThread(reactor, self._clear_queue)

        last_daq = 0
        while session.status in ['starting', 'running']:
            if PCU is None:
                try:
                    PCU = pcu.PCU(port=self.port)
                    self.log.info('Connected to PCU')
                    PCU.clear_buffer()
                    self.log.info('Cleared buffer')
                except (ConnectionRefusedError, serial.serialutil.SerialException):
                    self.log.error(
                        "Could not connect to PCU. "
                        "Retrying after 30 sec..."
                    )
                    time.sleep(30)
                    continue
            try:
                now = time.time()
                if now - last_daq > 5:
                    self._get_and_publish_data(PCU, session)
                    last_daq = now

                self._process_actions(PCU)
                session.degraded = False
                time.sleep(0.1)
            except serial.serialutil.SerialException:
                self.log.error(
                    "Decive reports readiness to read but returned no data. "
                    "Reconnect to PCU."
                )
                PCU.close()
                PCU = None
                session.degraded = True

        PCU.close()

    def _stop_main(self, session, params):
        """
        Stop acq process.
        """
        session.set_status('stopping')
        return True, 'Set main status to stopping'


def make_parser(parser=None):
    """
    Build the argument parser for the Agent. Allows sphinx to automatically build documentation
    baised on this function
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--port', type=str, help="Path to USB node for the PCU")
    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='HWPPCUAgent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)
    hwppcu_agent = HWPPCUAgent(agent,
                               port=args.port)
    agent.register_task('send_command', hwppcu_agent.send_command, blocking=False)
    agent.register_process(
        'main', hwppcu_agent.main, hwppcu_agent._stop_main, startup=True)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import numpy as np


def gen_turnaround(turnaround_method, t0, az0, el0, v0, turntime, az_flag, el_flag,
                   point_group_batch, second_leg_time=None, second_leg_velocity=0,
                   step_time=0.05):
    from .drivers import TrackPoint
    """
    Generates the trajectory of a turnaround given the initial position and velocity of the platform.
    This function generates a list of TrackPoints for the platform to follow during turnarounds, which
    is necessary for type2/3 scans.

    Args:
        turnaround_method (string): The method used for generating the turnaround. The current choices for turnaround
        methods are:
            1. "standard_gen": Generates a standard turnaround that mimics the trajectory of the spline based approach
                employed by the ACU with a set of TrackPoints that span the entirety of the turnaround.
            2. "three_leg": Generates a special turnaround that splits the turnaround into three "legs" that attempt to
                minimize the acceleration at the midpoint of the turnaround. See the _gen_three_leg_turnaround() docstring
                for further documentation.
            3. "two_leg": Generates a "three_leg" turnaround with second_leg_time = 0.
        t0 (float): The initial time of the turnaround.
        az0 (float): The iniital azimuth position of the turnaround. Should be equal to the final azimuth position of the turnaround.
        el0 (float): The initial elevation of the turnaround. El velocity is forced to 0 here so this is only used for creating TrackPoints.
        v0 (float): The initial azimuth velocity of the turnaround.
        turntime (float): The turnaround time given by the above equation.
        az_flag (int): The az flag used by the ACU. Inherited from the scan generation function and not changed. Used for TrackPoints.
        el_flag (int): The el flag used by the ACU. Inherited from the scan generation function and not changed. Used for TrackPoints.
        point_group_batch (int): the point group batch used by the ACU. Inherited from the scan generation function and not changed. Used for TrackPoints.
        second_leg_time (float): The time used by the second leg of the turnaround. Defaults to 1 second.
                               This limits the minimum turnaround time to ~2.0 seconds!
        second_leg_velocity (float): The velocity targeted by the beginning/end of the second leg of the turnaround. Defaults to 0 deg/s.
                                   second_leg acceleration = 2.0 * second_leg_velocity / second_leg_time.
        step_time (float): The step time between points in the turnaround. Defaults to 0.1 seconds (10Hz).

    Returns:
        turnaround_track (list of TrackPoints): A list of generated TrackPoints for the platform to follow. If the turnaround_method
        is not "standard_gen" or "three_leg" this function will return an empty list, which will cause the turnaround to follow the
        ACU's spline based approach track (i.e. a "standard" turnaround).
    """

    # Enforce 0 el velocity. Can be changed later if these want to be used with type2 or type3 LAT scans,
    # but more work is necessary to get to that point. We shouldn't mix the two yet!
    el_vel = 0.

    if turnaround_method == "standard_gen":
        ts, azs, vs = _gen_standard_turnaround(v0, turntime, step_time)

    elif turnaround_method == "two_leg":
        # Two leg turnarounds set second_leg_time = 0
        ts, azs, vs = _gen_three_leg_turnaround(v0, turntime, 0., second_leg_velocity, step_time)

    elif turnaround_method == "three_leg":
        ts, azs, vs = _gen_three_leg_turnaround(v0, turntime, second_leg_time, second_leg_velocity, step_time)

    else:  # Return an empty list for non generative turnaround methods such as "standard"
        return []

    ts += t0
    azs += az0

    # Turn our turnaround solution into TrackPoint's for the ACU.
    turnaround_track = []
    for t, az, v in zip(ts, azs, vs):
        turnaround_track.append(TrackPoint(timestamp=t,
                                az=az, el=el0, az_vel=v, el_vel=el_vel,
                                az_flag=az_flag, el_flag=el_flag,
                                group_flag=int(point_group_batch > 0)))

    return turnaround_track


def _gen_standard_turnaround(v0, turntime, step_time=0.05):
    """
    Function for generating the times, azimuths, and velocities for a standard turnaround.
    This function mimics the trajectory generated by the ACU in its spline based approach.

    Average acceleration is = 2 * v0/turntime.
    Peak acceleration is = 1.5*average_acceleration = 3 * v0/turn_time

    Args:
        v0 (float): The initial velocity at the start of the turnaround.
        turntime (float): The total time for the turnaround defined by 2*velocity/acceleration.
        step_time (float): The step time between points in the turnaround. Defaults to 0.05 seconds (20Hz).

    Returns:
        ts (float array): Array of times of the turnaround.
        azs (float array): Array of azimuth positions of the turnaround.
        vs (float array): Array of azimuth velocities of the turnaround.
    """

    t_start = 0  # We have to solve the trajectory around 0 or the linear equations become very large. Add t back on later
    t_target = t_start + turntime
    az_start = 0
    az_target = 0  # Turnaround should end at the same azimuth it started at.
    v_start = v0
    v_target = -1 * v0
    a_start = 0  # We always target an acceleration of 0 at the beginning and end of turnaround legs.
    a_target = 0
    ts, azs, vs = _gen_trajectory(t_start, t_target, 0,
                                  az_start, az_target, v_start,
                                  v_target, a_start, a_target,
                                  step_time, turnaround_method='standard')

    return ts, azs, vs


def _gen_three_leg_turnaround(v0, turntime, second_leg_time=None, second_leg_velocity=0, step_time=0.05):
    """
    Function for generating the times, azimuths, and velocities for a three_leg turnaround.
    This function generates a turnaround in three "legs":

        1. The initial deceleration.
        2. The middle leg with a low velocity/acceleration to gently turn the gears of the motors around so
           they contact the other face of the bearing with minimal force. The default velocity and acceleration
           is 0 so the platform comes to a full stop in this leg by a default.
        3. The final acceleration to the scan velocity in the opposite direction.

    The turnaround time of this function adheres to the same equation as the "baseline" turnaround function:

        turntime = (2.0 * scan_velocity) / scan_acceleration

    Thus, for the same scan velocity and scan acceleration this turnaround will take the same time as the baseline.

    The average and peak accelerations of the three_leg turnaround are as follows:

        Average acceleration is = 2 * v0/turntime.
        Peak acceleration is = 2.81*average_acceleration = 5.62 * v0/turn_time

    The peak acceleration of the three_leg turnaround is ~ 2.81 / 1.5 ~= 1.87x larger than the standard_gen turnaround.

    Args:
        v0 (float): The initial velocity at the start of the turnaround.
        turntime (float): The total time for the turnaround defined by 2*velocity/acceleration.
        second_leg_time (float): The time used by the second leg of the turnaround. Defaults to 1 second.
                               This limits the minimum turnaround time to ~2.0 seconds!
        second_leg_velocity (float): The velocity targeted by the beginning/end of the second leg of the turnaround. Defaults to 0 deg/s.
                                   second_leg acceleration = 2.0 * second_leg_velocity / second_leg_time.
        step_time (float): The step time between points in the turnaround. Defaults to 0.05 seconds (20Hz).

    Returns:
        ts (float array): Array of times of the turnaround.
        azs (float array): Array of azimuth positions of the turnaround.
        vs (float array): Array of azimuth velocities of the turnaround.
    """

    if second_leg_time is None:
        second_leg_time = turntime / 3.0  # Cut the turnaround into equal thirds unless otherwise specified.
    if second_leg_time == 0:
        assert (second_leg_velocity == 0)
        second_leg_acceleration = 0.
    else:
        second_leg_acceleration = 2.0 * second_leg_velocity / second_leg_time

    # Assert we have at least 0.5 seconds for the first and second legs of the turnaround!
    # This limits the turntime to >= 1.5 seconds
    assert (turntime - second_leg_time) >= 1.0, \
        "Time for the second leg of the turnaround is too long! The time remaining for the first and third legs is < 1.0 seconds!"

    # Solve for the first leg of the turnaround
    t_start_1 = 0  # We have to solve the trajectory around 0 or the linear equations become very large. Add t back on later
    t_target_1 = t_start_1 + (turntime - second_leg_time) / 2  # The first and third legs share the same portion of the turnaround time.
    az_start_1 = 0
    v_start_1 = v0
    v_target_1 = second_leg_velocity * np.sign(v_start_1)
    a_start_1 = 0
    a_target_1 = second_leg_acceleration * -1 * np.sign(v0)
    j_start_1 = 0  # We always target a jerk of 0 at the beginning and end of turnaround legs.
    j_target_1 = 0
    ts_1, azs_1, vs_1 = _gen_trajectory(t_start_1, t_target_1, az_start_1,
                                        v_start_1, v_target_1, a_start_1,
                                        a_target_1, j_start_1, j_target_1,
                                        step_time, turnaround_method='three_leg')

    # Solve for the second leg of the turnaround
    t_start_2 = t_target_1
    t_target_2 = t_start_2 + second_leg_time
    az_start_2 = azs_1[-1]  # The acceleration of the beggining of the next turnaround leg should always match the end of the last leg.
    v_start_2 = v_target_1
    v_target_2 = v_start_2 * -1
    a_start_2 = a_target_1
    a_target_2 = a_start_2
    j_start_2 = 0
    j_target_2 = 0
    if second_leg_time == 0:
        ts_2, azs_2, vs_2 = [t_start_2, t_target_2], [az_start_2, az_start_2], [0., 0.]
    else:
        ts_2, azs_2, vs_2 = _gen_trajectory(t_start_2, t_target_2, az_start_2,
                                            v_start_2, v_target_2, a_start_2,
                                            a_target_2, j_start_2, j_target_2,
                                            step_time, turnaround_method='three_leg')

    # Solve for the third leg of the turnaround
    t_start_3 = t_target_2
    t_target_3 = t_start_3 + (turntime - second_leg_time) / 2.0
    az_start_3 = azs_2[-1]
    v_start_3 = v_target_2
    v_target_3 = v0 * -1
    a_start_3 = a_target_2
    a_target_3 = 0
    j_start_3 = 0
    j_target_3 = 0
    ts_3, azs_3, vs_3 = _gen_trajectory(t_start_3, t_target_3, az_start_3,
                                        v_start_3, v_target_3, a_start_3,
                                        a_target_3, j_start_3, j_target_3,
                                        step_time, turnaround_method='three_leg')

    # Concatenate the times, azimuth positions, and azimuth velocities together.
    # The first point of each leg is a duplicate of the last so we drop those points.
    ts = np.concatenate([ts_1[1:], ts_2[1:-1], ts_3[1:]])
    azs = np.concatenate([azs_1[1:], azs_2[1:-1], azs_3[1:]])
    vs = np.concatenate([vs_1[1:], vs_2[1:-1], vs_3[1:]])

    return ts, azs, vs


def _gen_trajectory(t_i, t_f, xn1_i, x0_i, x0_f, x1_i, x1_f, x2_i, x2_f, step_time, turnaround_method):
    """
    Generally, generates the trajectory that minimizes the third derivative of the parameter defined by x0.

    In the context of this module, this function is used to generate the legs of the 3part turnaround in a way that
    minimizes the snap of the motion. Because we don't know the final positions of each of the legs we must
    generate the trajectories using the initial and final velocity, acceleration, and jerk, which minimizes the snap.

    --------------------------------------------------------------------------------------------------------------------------

    When turnaround method is 'standard', x0 is the function of position, x1 is velocity, x2 is acceleration,
    and xn1 is undefined/unused.

    Args:
        t_i (float): The initial time of the trajectory.
        t_f (float): The final time of the trajectory.
        xn1_i (float): Not used.
        x0_i (float): The initial position.
        x0_f (float): The final position.
        x1_i (float): The initial velocity.
        x1_f (float): The final velocity.
        x2_i (float): The initial acceleration.
        x2_f (float): The final acceleration.

    --------------------------------------------------------------------------------------------------------------------------

    When turnaround_method is 'three_leg', x0 is the function of velocity, x1 is acceleration, x2 is jerk, and xn1 is position.

    Args:
        t_i (float): The initial time of the trajectory.
        t_f (float): The final time of the trajectory.
        xn1_i (float): The initial position.
        x0_i (float): The initial velocity.
        x0_f (float): The final velocity.
        x1_i (float): The initial acceleration.
        x1_f (float): The final acceleration.
        x2_i (float): The initial jerk.
        x2_f (float): The final jerk.

    --------------------------------------------------------------------------------------------------------------------------

    Returns:
        ts (float array): A numpy array of timestamps.
        xs (float array): A numpy array of azimuth positions.
        vs (float array): A numpy array of azimuth velocities.
    """

    # Solve for the polynomial components that fits our initial and final conditions
    A = solve_fifth_polynomial_lin_eqs(t_i, t_f, x0_i, x0_f, x1_i, x1_f, x2_i, x2_f)

    ts = np.arange(t_i, t_f + step_time, step_time)  # Divide our times into points with step_time spacing

    if turnaround_method == 'standard':
        xs = np.polyval(A[::-1], ts)
        vs = np.polyval(np.polyder(A[::-1]), ts)

    elif turnaround_method == 'three_leg':
        vs = np.polyval(A[::-1], ts)

        xs = np.polyval(np.polyint(A[::-1]), ts)
        xs = xs - xs[0] + xn1_i  # Solve for the positions of each point

    return ts, xs, vs


# Linear Algebra Below
def solve_fifth_polynomial_lin_eqs(t_i, t_f, x0_i, x0_f, x1_i, x1_f, x2_i, x2_f):
    """
    Solves for the components of a polynomial equation of order five the form:

        x0 = A0 + A1*x + A2*x^2 + A3*x^3 + A4*x^4 + A5*a^5,

    given the initial/final conditions of the 0th, 1st, and 2nd derivatives.

    This solution minimizes the third derivative over the trajectory between t_i and t_f.

    Args:
        t_i (float): starting time
        t_f (float): stop time
        x0_i (float): initial 0th derivative condition
        x0_f (float): final 0th derivative condition
        x1_i (float): initial 1st derivative condition
        x1_f (float): final 1st derivative condition
        x2_i (float): initial 2nd derivative condition
        x2_f (float): initial 2nd derivative condition

    Returns:
        A0 (float): The solved 0th parameter of the order five polynomial.
        A1 (float): The solved 1st parameter of the order five polynomial.
        A2 (float): The solved 2nd parameter of the order five polynomial.
        A3 (float): The solved 3rd parameter of the order five polynomial.
        A4 (float): The solved 4th parameter of the order five polynomial.
        A5 (float): The solved 5th parameter of the order five polynomial.
    """

    x0 = [1, 1, 1, 1, 1, 1]
    x1 = [0, 1, 2, 3, 4, 5]
    x2 = [0, 0, 2, 6, 12, 20]

    A = np.zeros((6, 6))
    for i, x in enumerate([x0, x1, x2]):
        for j, y in enumerate(x):
            if j < i:
                continue

            A[2 * i, j] = y * t_i**(j - i)
            A[2 * i + 1, j] = y * t_f**(j - i)

    B = np.zeros(6)
    B[0] = x0_i
    B[1] = x0_f
    B[2] = x1_i
    B[3] = x1_f
    B[4] = x2_i
    B[5] = x2_f

    return np.linalg.solve(A, B)
